# Copyright (C) 2007 Alexandre Conrad, alexandre (dot) conrad (at) gmail (dot) com
#
# This module is part of FormAlchemy and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

import os
import cgi

#import logging
#logger = logging.getLogger('formalchemy.' + __name__)

from copy import copy, deepcopy
import datetime
import warnings

from sqlalchemy.orm import class_mapper, Query
from sqlalchemy.orm.attributes import ScalarAttributeImpl, ScalarObjectAttributeImpl, CollectionAttributeImpl, InstrumentedAttribute
from sqlalchemy.orm.properties import CompositeProperty, ColumnProperty
from sqlalchemy.exceptions import InvalidRequestError # 0.4 support

# formalchemy imports
from formalchemy import helpers as h
from formalchemy import fatypes, validators
from formalchemy import config
from formalchemy.i18n import get_translator
from formalchemy.i18n import _

__all__ = ['Field', 'FieldRenderer',
           'TextFieldRenderer', 'TextAreaFieldRenderer',
           'PasswordFieldRenderer', 'HiddenFieldRenderer',
           'DateFieldRenderer', 'TimeFieldRenderer',
           'DateTimeFieldRenderer',
           'CheckBoxFieldRenderer', 'CheckBoxSet']

def iterable(item):
    try:
        iter(item)
    except:
        return False
    return True

def _stringify(k, null_value=''):
    if k is None:
        return null_value
    if isinstance(k, str):
        return unicode(k, config.encoding)
    elif isinstance(k, unicode):
        return k
    elif hasattr(k, '__unicode__'):
        return unicode(k)
    else:
        return unicode(str(k), config.encoding)

class FieldRenderer(object):
    """
    This should be the super class of all Renderer classes.

    Renderers generate the html corresponding to a single Field,
    and are also responsible for deserializing form data into
    Python objects.

    Subclasses should override `render` and `deserialize`.
    See their docstrings for details.
    """
    def __init__(self, field):
        self.field = field
        assert isinstance(self.field, AbstractField)

    def name(self):
        """Name of rendered input element"""
        clsname = self.field.model.__class__.__name__
        pk = self.field.parent._bound_pk
        assert pk != ''
        if isinstance(pk, basestring) or not iterable(pk):
            pk_string = _stringify(pk)
        else:
            # remember to use a delimiter that can be used in the DOM (specifically, no commas).
            # we don't have to worry about escaping the delimiter, since we never try to
            # deserialize the generated name.  All we care about is generating unique
            # names for a given model's domain.
            pk_string = u'_'.join([_stringify(k) for k in pk])

        components = [clsname, pk_string, self.field.name]
        if self.field.parent.prefix:
            components.insert(0, self.field.parent.prefix)
        return u"-".join(components)
    name = property(name)

    def _value(self):
        """
        Submitted value, or field value converted to string.
        Return value is always either None or a string.
        """
        if not self.field.is_readonly() and self._params is not None:
            # submitted value.  do not deserialize here since that requires valid data, which we might not have
            v = self._serialized_value() 
        else:
            v = None
        # empty field will be '' -- use default value there, too
        return v or self._model_value_as_string()
    _value = property(_value)

    def _model_value_as_string(self):
        if self.field.model_value is None:
            return None
        if self.field.is_collection:
            return [self.stringify_value(v) for v in self.field.model_value]
        else:
            return self.stringify_value(self.field.model_value)

    def get_translator(self, **kwargs):
        """return a GNUTranslations object in the most convenient way
        """
        if 'F_' in kwargs:
            return kwargs.pop('F_')
        if 'lang' in kwargs:
            lang = kwargs.pop('lang')
        else:
            lang = 'en'
        return get_translator(lang=lang).gettext

    def render(self, **kwargs):
        """
        Render the field.  Use `self.name` to get a unique name for the
        input element and id.  `self._value` may also be useful if
        you are not rendering multiple input elements.
        """
        raise NotImplementedError()

    def render_readonly(self, **kwargs):
        """render a string representation of the field value"""
        value = self.field.raw_value
        if value is None:
            return ''
        if self.field.is_scalar_relation:
            q = self.field.query(self.field.relation_type())
            v = q.get(value)
            return _stringify(v)
        if isinstance(value, list):
            return u', '.join([_stringify(item) for item in value])
        if isinstance(value, unicode):
            return value
        return _stringify(value)

    def _params(self):
        return self.field.parent.data
    _params = property(_params)

    def _serialized_value(self):
        """
        Returns the appropriate value to deserialize for field's
        datatype, from the user-submitted data.  Only called
        internally, so, if you are overriding `deserialize`,
        you can use or ignore `_serialized_value` as you please.

        This is broken out into a separate method so multi-input
        renderers can stitch their values back into a single one
        to have that can be handled by the default deserialize.

        Do not attempt to deserialize here; return value should be a
        string (corresponding to the output of `str` for your data
        type), or for a collection type, a a list of strings,
        or None if no value was submitted for this renderer.

        The default _serialized_value returns the submitted value(s)
        in the input element corresponding to self.name.
        """
        if self.field.is_collection:
            return self._params.getall(self.name)
        return self._params.getone(self.name)

    def deserialize(self):
        """
        Turns the user-submitted data into a Python value.  (The raw
        data will be available in self.field.parent.data, or you
        can use `_serialized_value` if it is convenient.)  For SQLAlchemy
        collections, return a list of primary keys, and !FormAlchemy
        will take care of turning that into a list of objects.
        For manually added collections, return a list of values.

        You should only have to override this if you are using custom
        (e.g., Composite) types.
        """
        if self.field.is_collection:
            return [self._deserialize(subdata) for subdata in self._serialized_value()]
        return self._deserialize(self._serialized_value())

    def _deserialize(self, data):
        if isinstance(self.field.type, fatypes.Boolean):
            if data is not None:
                if data.lower() in ['1', 't', 'true', 'yes']: return True
                if data.lower() in ['0', 'f', 'false', 'no']: return False
        if data is None or data == self.field._null_option[1]:
            return None
        if isinstance(self.field.type, fatypes.Integer):
            return validators.integer(data, self)
        if isinstance(self.field.type, fatypes.Float):
            return validators.float_(data, self)
        if isinstance(self.field.type, fatypes.Numeric):
            if self.field.type.asdecimal:
                return validators.decimal_(data, self)
            else:
                return validators.float_(data, self)

        def _date(data):
            if data == 'YYYY-MM-DD' or data == '-MM-DD' or not data.strip():
                return None
            try:
                return datetime.date(*[int(st) for st in data.split('-')])
            except:
                raise validators.ValidationError('Invalid date')
        def _time(data):
            if data == 'HH:MM:SS' or not data.strip():
                return None
            try:
                return datetime.time(*[int(st) for st in data.split(':')])
            except:
                raise validators.ValidationError('Invalid time')

        if isinstance(self.field.type, fatypes.Date):
            return _date(data)
        if isinstance(self.field.type, fatypes.Time):
            return _time(data)
        if isinstance(self.field.type, fatypes.DateTime):
            data_date, data_time = data.split(' ')
            dt, tm = _date(data_date), _time(data_time)
            if dt is None and tm is None:
                return None
            elif dt is None or tm is None:
                raise validators.ValidationError('Incomplete datetime')
            return datetime.datetime(dt.year, dt.month, dt.day, tm.hour, tm.minute, tm.second)

        return data
    def stringify_value(self, v):
        return _stringify(v, null_value=self.field._null_option[1])

class EscapingReadonlyRenderer(FieldRenderer):
    """
    In readonly mode, html-escapes the output of the default renderer
    for this field type.  (Escaping is not performed by default because
    it is sometimes useful to have the renderer include raw html in its
    output.  The FormAlchemy admin app extension for Pylons uses this,
    for instance.)
    """
    def __init__(self, field):
        FieldRenderer.__init__(self, field)
        self._renderer = field._get_renderer()(field)

    def render(self, **kwargs):
        return self._renderer.render(**kwargs)

    def render_readonly(self, **kwargs):
        return h.html_escape(self._renderer.render_readonly(**kwargs))


class TextFieldRenderer(FieldRenderer):
    """render a field as a text field"""
    def length(self):
        return self.field.type.length
    length = property(length)

    def render(self, **kwargs):
        return h.text_field(self.name, value=self._value, maxlength=self.length, **kwargs)


class IntegerFieldRenderer(FieldRenderer):
    """render an integer as a text field"""
    def render(self, **kwargs):
        return h.text_field(self.name, value=self._value, **kwargs)


class FloatFieldRenderer(FieldRenderer):
    """render a float as a text field"""
    def render(self, **kwargs):
        return h.text_field(self.name, value=self._value, **kwargs)


class PasswordFieldRenderer(TextFieldRenderer):
    """Render a password field"""
    def render(self, **kwargs):
        return h.password_field(self.name, value=self._value, maxlength=self.length, **kwargs)
    def render_readonly(self):
        return '*'*6

class TextAreaFieldRenderer(FieldRenderer):
    """render a field as a textarea"""
    def render(self, **kwargs):
        if isinstance(kwargs.get('size'), tuple):
            kwargs['size'] = 'x'.join([str(i) for i in kwargs['size']])
        return h.text_area(self.name, content=self._value, **kwargs)


class HiddenFieldRenderer(FieldRenderer):
    """render a field as an hidden field"""
    def render(self, **kwargs):
        return h.hidden_field(self.name, value=self._value, **kwargs)
    def render_readonly(self):
        return ''


class CheckBoxFieldRenderer(FieldRenderer):
    """render a boolean value as checkbox field"""
    def render(self, **kwargs):
        return h.check_box(self.name, True, checked=_simple_eval(self._value or ''), **kwargs)
    def _serialized_value(self):
        if self.name not in self._params:
            return None
        return FieldRenderer._serialized_value(self)
    def deserialize(self):
        if self._serialized_value() is None:
            return False
        return FieldRenderer.deserialize(self)

class FileFieldRenderer(FieldRenderer):
    """render a file input field"""
    remove_label = _('Remove')
    def __init__(self, *args, **kwargs):
        FieldRenderer.__init__(self, *args, **kwargs)
        self._data = None # caches FieldStorage data
        self._filename = None

    def render(self, **kwargs):
        if self.field.model_value:
            checkbox_name = '%s--remove' % self.name
            return '%s %s %s' % (
                   h.file_field(self.name, **kwargs),
                   h.check_box(checkbox_name),
                   h.label(self.remove_label, for_=checkbox_name))
        else:
            return h.file_field(self.name, **kwargs)

    def get_size(self):
        value = self.field.raw_value
        if value is None:
            return 0
        return len(value)

    def readable_size(self):
        length = self.get_size()
        if length == 0:
            return '0 KB'
        if length <= 1024:
            return '1 KB'
        if length > 1048576:
            return '%0.02f MB' % (length / 1048576.0)
        return '%0.02f KB' % (length / 1024.0)

    def render_readonly(self, **kwargs):
        """
        render only the binary size in a human readable format but you can
        override it to whatever you want
        """
        return self.readable_size()

    def deserialize(self):
        data = FieldRenderer.deserialize(self)
        if isinstance(data, cgi.FieldStorage):
            if data.filename:
                # FieldStorage can only be read once so we need to cache the
                # value since FA call deserialize during validation and
                # synchronisation
                if self._data is None:
                    self._filename = data.filename
                    self._data = data.file.read()
                data = self._data
            else:
                data = None
        checkbox_name = '%s--remove' % self.name
        if not data and not self._params.has_key(checkbox_name):
            data = getattr(self.field.model, self.field.name)
        return data is not None and data or ''

# for when and/or is not safe b/c first might eval to false
def _ternary(condition, first, second):
    if condition:
        return first()
    return second()

class DateFieldRenderer(FieldRenderer):
    """Render a date field"""
    format = '%Y-%m-%d'
    edit_format = 'm-d-y'
    def render_readonly(self, **kwargs):
        value = self.field.raw_value
        return value and value.strftime(self.format) or ''
    def _render(self, **kwargs):
        data = self._params
        F_ = self.get_translator(**kwargs)
        month_options = [(F_('Month'), 'MM')] + [(F_('month_%02i' % i), str(i)) for i in xrange(1, 13)]
        day_options = [(F_('Day'), 'DD')] + [(i, str(i)) for i in xrange(1, 32)]
        mm_name = self.name + '__month'
        dd_name = self.name + '__day'
        yyyy_name = self.name + '__year'
        mm = _ternary((data is not None and mm_name in data), lambda: data[mm_name],  lambda: str(self.field.model_value and self.field.model_value.month))
        dd = _ternary((data is not None and dd_name in data), lambda: data[dd_name], lambda: str(self.field.model_value and self.field.model_value.day))
        # could be blank so don't use and/or construct
        if data is not None and yyyy_name in data:
            yyyy = data[yyyy_name]
        else:
            yyyy = str(self.field.model_value and self.field.model_value.year or 'YYYY')
        selects = dict(
                m=h.select(mm_name, h.options_for_select(month_options, selected=mm), **kwargs),
                d=h.select(dd_name, h.options_for_select(day_options, selected=dd), **kwargs),
                y=h.text_field(yyyy_name, value=yyyy, maxlength=4, size=4, **kwargs))
        value = [selects.get(l) for l in self.edit_format.split('-')]
        return ' '.join(value)
    def render(self, **kwargs):
        return h.content_tag('span', self._render(**kwargs), id=self.name)

    def _serialized_value(self):
        return '-'.join([self._params.getone(self.name + '__' + subfield) for subfield in ['year', 'month', 'day']])


class TimeFieldRenderer(FieldRenderer):
    """Render a time field"""
    format = '%H:%M:%S'
    def render_readonly(self, **kwargs):
        value = self.field.raw_value
        return value and value.strftime(self.format) or ''
    def _render(self, **kwargs):
        data = self._params
        hour_options = ['HH'] + [(i, str(i)) for i in xrange(24)]
        minute_options = ['MM' ] + [(i, str(i)) for i in xrange(60)]
        second_options = ['SS'] + [(i, str(i)) for i in xrange(60)]
        hh_name = self.name + '__hour'
        mm_name = self.name + '__minute'
        ss_name = self.name + '__second'
        hh = _ternary((data is not None and hh_name in data), lambda: data[hh_name], lambda: str(self.field.model_value and self.field.model_value.hour))
        mm = _ternary((data is not None and mm_name in data), lambda: data[mm_name], lambda: str(self.field.model_value and self.field.model_value.minute))
        ss = _ternary((data is not None and ss_name in data), lambda: data[ss_name], lambda: str(self.field.model_value and self.field.model_value.second))
        return h.select(hh_name, h.options_for_select(hour_options, selected=hh), **kwargs) \
               + ':' + h.select(mm_name, h.options_for_select(minute_options, selected=mm), **kwargs) \
               + ':' + h.select(ss_name, h.options_for_select(second_options, selected=ss), **kwargs)
    def render(self, **kwargs):
        return h.content_tag('span', self._render(**kwargs), id=self.name)

    def _serialized_value(self):
        return ':'.join([self._params.getone(self.name + '__' + subfield) for subfield in ['hour', 'minute', 'second']])


class DateTimeFieldRenderer(DateFieldRenderer, TimeFieldRenderer):
    """Render a date time field"""
    format = '%Y-%m-%d %H:%M:%S'
    def render(self, **kwargs):
        return h.content_tag('span', DateFieldRenderer._render(self, **kwargs) + ' ' + TimeFieldRenderer._render(self, **kwargs), id=self.name)

    def _serialized_value(self):
        return DateFieldRenderer._serialized_value(self) + ' ' + TimeFieldRenderer._serialized_value(self)


def _extract_options(options):
    if isinstance(options, dict):
        options = options.items()
    for choice in options:
        # Choice is a list/tuple...
        if isinstance(choice, (list, tuple)):
            if len(choice) != 2:
                raise Exception('Options should consist of two items, a name and a value; found %d items in %r' % (len(choice, choice)))
            yield choice
        # ... or just a string.
        else:
            if not isinstance(choice, basestring):
                raise Exception('List, tuple, or string value expected as option (got %r)' % choice)
            yield (choice, choice)


class RadioSet(FieldRenderer):
    """render a field as radio"""
    widget = staticmethod(h.radio_button)
    format = '%(field)s%(label)s'
    
    def _serialized_value(self):
        if self.name not in self._params:
            return None
        return FieldRenderer._serialized_value(self)

    def _is_checked(self, choice_value):
        return self._value == _stringify(choice_value)

    def render(self, options, **kwargs):
        self.radios = []
        for i, (choice_name, choice_value) in enumerate(_extract_options(options)):
            choice_id = '%s_%i' % (self.name, i)
            radio = self.widget(self.name, choice_value, id=choice_id,
                                checked=self._is_checked(choice_value), **kwargs)
            label = h.content_tag('label', choice_name, for_=choice_id)
            self.radios.append(self.format % dict(field=radio,
                                                  label=label))
        return h.tag("br").join(self.radios)


class CheckBoxSet(RadioSet):
    widget = staticmethod(h.check_box)

    def _serialized_value(self):
        if self.name not in self._params:
            return []
        return FieldRenderer._serialized_value(self)

    def _is_checked(self, choice_value):
        return _stringify(choice_value) in self._value


class SelectFieldRenderer(FieldRenderer):
    """render a field as select"""
    def _serialized_value(self):
        if self.name not in self._params:
            if self.field.is_collection:
                return []
            return None
        return FieldRenderer._serialized_value(self)

    def render(self, options, **kwargs):
        if callable(options):
            L = _normalized_options(options(self.field.parent))
            if not self.field.is_required() and not self.field.is_collection:
                L.insert(0, self.field._null_option)
        else:
            L = list(options)
        if len(L) > 0:
            if len(L[0]) == 2:
                L = [(k, self.stringify_value(v)) for k, v in L]
            else:
                L = [_stringify(k) for k in L]
        return h.select(self.name, h.options_for_select(L, selected=self._value), **kwargs)

    def render_readonly(self, options=None, **kwargs):
        """render a string representation of the field value.
           Try to retrieve a value from `options`
        """
        if not options or self.field.is_scalar_relation:
            return FieldRenderer.render_readonly(self)

        value = self.field.raw_value
        if value is None:
            return ''

        if callable(options):
            L = _normalized_options(options(self.field.parent))
        else:
            L = list(options)

        if len(L) > 0:
            if len(L[0]) == 2:
                L = [(v, k) for k, v in L]
            else:
                L = [(k, _stringify(k)) for k in L]
        D = dict(L)
        if isinstance(value, list):
            return u', '.join([_stringify(D.get(item, item)) for item in value])
        return _stringify(D.get(value, value))


def _pk_one_column(instance, column):
    try:
        attr = getattr(instance, column.key)
    except AttributeError:
        # FIXME: this is not clean but the only way i've found to retrieve the
        # real attribute name of the primary key.
        # This is needed when you use something like:
        #    id = Column('UGLY_NAMED_ID', primary_key=True)
        # It's a *really* needed feature
        cls = instance.__class__
        for k in instance._sa_class_manager.keys():
            props = getattr(cls, k).property
            if hasattr(props, 'columns'):
                if props.columns[0] is column:
                    attr = getattr(instance, k)
                    break
    return attr

def _pk(instance):
    # Return the value of this instance's primary key, suitable for passing to Query.get().  
    # Will be a tuple if PK is multicolumn.
    try:
        columns = class_mapper(type(instance)).primary_key
    except InvalidRequestError:
        return None
    if len(columns) == 1:
        return _pk_one_column(instance, columns[0])
    return tuple([_pk_one_column(instance, column) for column in columns])


# see http://code.activestate.com/recipes/364469/ for explanation.
# 2.6 provides ast.literal_eval, but requiring 2.6 is a bit of a stretch for now.
import compiler
class _SafeEval(object):
    def visit(self, node,**kw):
        cls = node.__class__
        meth = getattr(self,'visit'+cls.__name__,self.default)
        return meth(node, **kw)
            
    def default(self, node, **kw):
        for child in node.getChildNodes():
            return self.visit(child, **kw)
            
    visitExpression = default
    
    def visitName(self, node, **kw):
        if node.name in ['True', 'False', 'None']:
            return eval(node.name)

    def visitConst(self, node, **kw):
        return node.value

    def visitTuple(self,node, **kw):
        return tuple(self.visit(i) for i in node.nodes)
        
    def visitList(self,node, **kw):
        return [self.visit(i) for i in node.nodes]

def _simple_eval(source):
    """like 2.6's ast.literal_eval, but only does constants, lists, and tuples, for serialized pk eval"""
    if source == '':
        return None
    walker = _SafeEval()
    ast = compiler.parse(source, 'eval')
    return walker.visit(ast)


def _query_options(L):
    """
    Return a list of tuples of `(item description, item pk)`
    for each item in the iterable L, where `item description`
    is the result of str(item) and `item pk` is the item's primary key.
    """
    return [(_stringify(item), _pk(item)) for item in L]


def _normalized_options(options):
    """
    If `options` is an SA query or an iterable of SA instances, it will be
    turned into a list of `(item description, item value)` pairs. Otherwise, a
    copy of the original options will be returned with no further validation.
    """
    if isinstance(options, Query):
        options = options.all()
    if callable(options):
        return options
    i = iter(options)
    try:
        first = i.next()
    except StopIteration:
        return []
    try:
        class_mapper(type(first))
    except:
        return list(options)
    return _query_options(options)


def _foreign_keys(property):
    # 0.4/0.5 compatibility fn
    try:
        return property.foreign_keys
    except AttributeError:
        return [r for l, r in property.synchronize_pairs]


def _model_equal(a, b):
    if not isinstance(a, type):
        a = type(a)
    if not isinstance(b, type):
        b = type(b)
    return a is b


class AbstractField(object):
    """
    Contains the information necessary to render (and modify the rendering of)
    a form field
    """
    _null_option = (u'None', u'')

    def __init__(self, parent):
        # the FieldSet (or any ModelRenderer) owning this instance
        self.parent = parent
        if 0:
            import forms
            isinstance(self.parent, forms.FieldSet)
        # Renderer for this Field.  this will
        # be autoguessed, unless the user forces it with .dropdown,
        # .checkbox, etc.
        self._renderer = None
        # other render options, such as size, multiple, etc.
        self.render_opts = {}
        # validator functions added with .validate()
        self.validators = []
        # errors found by _validate() (which runs implicit and
        # explicit validators)
        self.errors = []
        self._readonly = False
        # label to use for the rendered field.  autoguessed if not specified by .label()
        self.label_text = None
        # optional attributes to pass to renderers
        self.html_options = {}
        # True iff this Field is a primary key
        self.is_pk = False
        # True iff this Field is a raw foreign key
        self.is_raw_foreign_key = False
        # Field metadata, for customization
        self.metadata = {}
        return False

    def __deepcopy__(self, memo):
        wrapper = copy(self)
        wrapper.render_opts = dict(self.render_opts)
        wrapper.validators = list(self.validators)
        wrapper.errors = list(self.errors)
        try:
            wrapper._renderer = copy(self._renderer)
        except TypeError: # 2.4 support
            # it's a lambda, safe to just use same referende
            pass
        if hasattr(wrapper._renderer, 'field'):
            wrapper._renderer.field = wrapper
        return wrapper

    def requires_label(self):
        return not isinstance(self.renderer, HiddenFieldRenderer)
    requires_label = property(requires_label)

    def query(self, *args, **kwargs):
        """Perform a query in the parent's session"""
        if not self.parent.session:
            raise Exception("No session found.  Either bind a session explicitly, or specify relation options manually so FormAlchemy doesn't try to autoload them.")
        return self.parent.session.query(*args, **kwargs)

    def _validate(self):
        if self.is_readonly():
            return True

        self.errors = []

        try:
            value = self._deserialize()
        except validators.ValidationError, e:
            self.errors.append(e)
            return False

        L = list(self.validators)
        if self.is_required() and validators.required not in L:
            L.append(validators.required)
        for validator in L:
            if validator is not validators.required and value is None:
                continue
            try:
                validator(value, self)
            except validators.ValidationError, e:
                self.errors.append(e.message)
            except TypeError:
                warnings.warn(DeprecationWarning('Please provide a field argument to your %r validator. Your validator will break in FA 1.5' % validator))
                try:
                    validator(value)
                except validators.ValidationError, e:
                    self.errors.append(e.message)
        return not self.errors

    def is_required(self):
        """True iff this Field must be given a non-empty value"""
        return validators.required in self.validators

    def is_readonly(self):
        """True iff this Field is in readonly mode"""
        return self._readonly

    def model(self):
        return self.parent.model
    model = property(model)

    def _modified(self, **kwattrs):
        # return a copy of self, with the given attributes modified
        copied = deepcopy(self)
        for attr, value in kwattrs.iteritems():
            setattr(copied, attr, value)
        return copied
    def with_null_as(self, option):
        """Render null as the given option tuple of text, value."""
        return self._modified(_null_option=option)
    def with_renderer(self, renderer):
        """
        Return a copy of this Field, with a different renderer.
        Used for one-off renderer changes; if you want to change the
        renderer for all instances of a Field type, modify
        FieldSet.default_renderers instead.
        """
        return self._modified(_renderer=renderer)
    def bind(self, parent):
        """Return a copy of this Field, bound to a different parent"""
        return self._modified(parent=parent)
    def with_metadata(self, **attrs):
        """Attach some metadata attributes to the Field, to be used by
        conditions in templates.

        Example usage:

          >>> test = Field('test')
          >>> field = test.with_metadata(instructions='use this widget this way')
          ...

        And further in your templates you can verify:

          >>> 'instructions' in field.metadata
          True

        and display the content in a <span> or something.
        """
        new_attr = self.metadata.copy()
        new_attr.update(attrs)
        return self._modified(metadata=new_attr)
    def validate(self, validator):
        """
        Add the `validator` function to the list of validation
        routines to run when the `FieldSet`'s `validate` method is
        run. Validator functions take one parameter: the value to
        validate. This value will have already been turned into the
        appropriate data type for the given `Field` (string, int, float,
        etc.). It should raise `ValidationError` if validation
        fails with a message explaining the cause of failure.
        """
        field = deepcopy(self)
        field.validators.append(validator)
        return field
    def required(self):
        """
        Convenience method for `validate(validators.required)`. By
        default, NOT NULL columns are required. You can only add
        required-ness, not remove it.
        """
        return self.validate(validators.required)
    def with_html(self, **html_options):
        """
        Give some HTML options to renderer.

        Trailing underscore (_) characters will be stripped. For example,
        you might want to add a `class` attribute to your checkbox. You
        would need to specify `.options(class_='someclass')`.

        For WebHelpers-aware people: those parameters will be passed to
        the `text_area()`, `password()`, `text()`, etc.. webhelpers.

        NOTE: Those options can override generated attributes and can mess
              the `sync` calls, or `label`-tag associations (if you change
              `name`, or `id` for example).  Use with caution.
        """
        new_opts = copy(self.html_options)
        for k, v in html_options.iteritems():
            new_opts[k.rstrip('_')] = v
        return self._modified(html_options=new_opts)
    def label(self, text):
        """
        Change the label associated with this field.  By default, the field name
        is used, modified for readability (e.g., 'user_name' -> 'User name').
        """
        return self._modified(label_text=text)
    def readonly(self, value=True):
        """Render the field readonly."""
        return self._modified(_readonly=True)
    def hidden(self):
        """Render the field hidden.  (Value only, no label.)"""
        return self._modified(_renderer=HiddenFieldRenderer, render_opts={})
    def password(self):
        """Render the field as a password input, hiding its value."""
        field = deepcopy(self)
        field._renderer = lambda: field.parent.default_renderers['password']
        field.render_opts = {}
        return field
    def textarea(self, size=None):
        """
        Render the field as a textarea.  Size must be a string
        (`"25x10"`) or tuple (`25, 10`).
        """
        field = deepcopy(self)
        field._renderer = lambda: field.parent.default_renderers['textarea']
        if size:
            field.render_opts = {'size': size}
        return field
    def radio(self, options=None):
        """Render the field as a set of radio buttons."""
        field = deepcopy(self)
        field._renderer = lambda: field.parent.default_renderers['radio']
        if options is None:
            options = self.render_opts.get('options')
        else:
            options = _normalized_options(options)
        field.render_opts = {'options': options}
        return field
    def checkbox(self, options=None):
        """Render the field as a set of checkboxes."""
        field = deepcopy(self)
        field._renderer = lambda: field.parent.default_renderers['checkbox']
        if options is None:
            options = self.render_opts.get('options')
        else:
            options = _normalized_options(options)
        field.render_opts = {'options': options}
        return field
    def dropdown(self, options=None, multiple=False, size=5):
        """
        Render the field as an HTML select field.
        (With the `multiple` option this is not really a 'dropdown'.)
        """
        field = deepcopy(self)
        field._renderer = lambda: field.parent.default_renderers['dropdown']
        if options is None:
            options = self.render_opts.get('options')
        else:
            options = _normalized_options(options)
        field.render_opts = {'multiple': multiple, 'options': options}
        if multiple:
            field.render_opts['size'] = size
        return field
    def reset(self):
        """
        Return the field with all configuration changes reverted.
        """
        return deepcopy(self.parent._fields[self.name])

    def _get_renderer(self):
        for t in self.parent.default_renderers:
            if not isinstance(t, basestring) and isinstance(self.type, t):
                return self.parent.default_renderers[t]
        raise TypeError(
                'No renderer found for field %s. '
                'Type %s as no default renderer' % (self.name, self.type))

    def renderer(self):
        if self._renderer is None:
            self._renderer = self._get_renderer()
        if callable(self._renderer):
            # invoke potential lambda
            try:
                self._renderer = self._renderer()
            except TypeError:
                pass
        if callable(self._renderer):
            # must be a Renderer class.  instantiate.
            self._renderer = self._renderer(self)
        return self._renderer
    renderer = property(renderer)

    def _get_render_opts(self):
        """
        Calculate the final options dict to be sent to renderers.
        """
        # Use options from internally set render_opts
        opts = dict(self.render_opts)
        # Override with user-specified options (with .with_html())
        opts.update(self.html_options)
        return opts

    def render(self):
        """
        Render this Field as HTML.
        """
        if self.is_readonly():
            return self.render_readonly()

        opts = self._get_render_opts()

        if (isinstance(self.type, fatypes.Boolean)
            and not opts.get('options')
            and self.renderer.__class__ in [self.parent.default_renderers['dropdown'], self.parent.default_renderers['radio']]):
            opts['options'] = [('Yes', True), ('No', False)]
        return self.renderer.render(**opts)

    def render_readonly(self):
        """
        Render this Field as HTML for read only mode.
        """
        return self.renderer.render_readonly(**self._get_render_opts())

    def _pkify(self, value):
        """return the PK for value, if applicable"""
        return value

    def value(self):
        """
        The value of this Field: use the corresponding value in the bound `data`,
        if any; otherwise, use the value in the bound `model`.  For SQLAlchemy models,
        if there is still no value, use the default defined on the corresponding `Column`.

        For SQLAlchemy collections,
        a list of the primary key values of the items in the collection is returned.

        Invalid form data will cause an error to be raised.  Controllers should thus validate first.
        Renderers should thus never access .value; use .model_value instead.  
        """
        # TODO add ._validated flag to save users from themselves?
        if not self.is_readonly() and self.parent.data is not None:
            v = self._deserialize()
            if v is not None:
                return self._pkify(v)
        return self.model_value
    value = property(value)

    def model_value(self):
        """
        raw value from model, transformed if necessary for use as a form input value.
        """
        raise NotImplementedError()
        
    def raw_value(self):
        """
        raw value from model.  different from `.model_value` in SQLAlchemy fields, because for reference types,
        `.model_value` will return the foreign key ID.  This will return the actual object 
        referenced instead.
        """
        raise NotImplementedError()

    def _deserialize(self):
        return self.renderer.deserialize()

class Field(AbstractField):
    """
    A manually-added form field
    """
    def __init__(self, name=None, type=fatypes.String, value=None):
        """
          * `name`: field name
          * `type`: data type, from formalchemy.types (Boolean, Integer, String, etc.),
            or a custom type for which you have added a renderer.
          * `value`: default value.  If value is a callable, it will be passed
            the current bound model instance when the value is read.  This allows
            creating a Field whose value depends on the model once, then
            binding different instances to it later.
        """
        AbstractField.__init__(self, None) # parent will be set by ModelRenderer.add
        self.type = type()
        self.name = self.key = name
        self._value = value
        self.is_relation = False
        self.is_scalar_relation = False

    def model_value(self):
        return self.raw_value
    model_value = property(model_value)

    def is_collection(self):
        return self.render_opts.get('multiple', False) or isinstance(self.renderer, self.parent.default_renderers['checkbox'])
    is_collection = property(is_collection)

    def raw_value(self):
        try:
            # this is NOT the same as getattr -- getattr will return the class's
            # value for the attribute name, which for a manually added Field will
            # be the Field object.  So force looking in the instance __dict__ only.
            return self.model.__dict__[self.name]
        except KeyError:
            pass
        if callable(self._value):
            return self._value(self.model)
        return self._value
    raw_value = property(raw_value)

    def sync(self):
        """Set the attribute's value in `model` to the value given in `data`"""
        if not self.is_readonly():
            self._value = self._deserialize()

    def __repr__(self):
        return 'AttributeField(%s)' % self.name

    def __eq__(self, other):
        # we override eq so that when we configure with options=[...], we can match the renders in options
        # with the ones that were generated at FieldSet creation time
        try:
            return self.name == other.name and _model_equal(self.model, other.model)
        except (AttributeError, ValueError):
            return False
    def __hash__(self):
        return hash(self.name)


class AttributeField(AbstractField):
    """
    Field corresponding to an SQLAlchemy attribute.
    """
    def __init__(self, instrumented_attribute, parent):
        """
            >>> from formalchemy.tests import FieldSet, Order
            >>> fs = FieldSet(Order)
            >>> print fs.user.key
            user

            >>> print fs.user.name
            user_id
        """
        AbstractField.__init__(self, parent)
        # we rip out just the parts we care about from InstrumentedAttribute.
        # impl is the AttributeImpl.  So far all we care about there is ".key,"
        # which is the name of the attribute in the mapped class.
        self._impl = instrumented_attribute.impl
        # property is the PropertyLoader which handles all the interesting stuff.
        # mapper, columns, and foreign keys are all located there.
        self._property = instrumented_attribute.property

        # True iff this is a multi-valued (one-to-many or many-to-many) SA relation
        self.is_collection = isinstance(self._impl, CollectionAttributeImpl)

        # True iff this is the 'one' end of a one-to-many relation
        self.is_scalar_relation = isinstance(self._impl, ScalarObjectAttributeImpl)

        # True iff this field represents a mapped SA relation
        self.is_relation = self.is_scalar_relation or self.is_collection

        self.is_composite = isinstance(self._property, CompositeProperty)

        _columns = self._columns

        self.is_pk = bool([c for c in self._columns if c.primary_key])

        self.is_raw_foreign_key = bool(isinstance(self._property, ColumnProperty) and _foreign_keys(self._property.columns[0]))

        self.is_composite_foreign_key = len(_columns) > 1 and not [c for c in _columns if not _foreign_keys(c)]

        if self.is_composite:
            # this is a little confusing -- we need to return an _instance_ of
            # the correct type, which for composite values will be the value
            # itself. SA should probably have called .type something
            # different, or just not instantiated them...
            self.type = self._property.composite_class.__new__(self._property.composite_class)
        elif len(_columns) > 1:
            self.type = None # may have to be more accurate here
        else:
            self.type = _columns[0].type

        self.key = self._impl.key
        self._column_name = '_'.join([c.name for c in _columns])

        # The name of the form input. usually the same as the key, except for
        # single-valued SA relation properties. For example, for order.user,
        # name will be 'user_id' (assuming that is indeed the name of the foreign
        # key to users), but for user.orders, name will be 'orders'.
        if self.is_collection or self.is_composite or not hasattr(self.model, self._column_name):
            self.name = self.key
        else:
            self.name = self._column_name

        # smarter default "required" value
        if not self.is_collection and not self.is_readonly() and [c for c in _columns if not c.nullable]:
            self.validators.append(validators.required)

    def is_readonly(self):
        from sqlalchemy.sql.expression import _Label
        return AbstractField.is_readonly(self) or isinstance(self._columns[0], _Label)

    def _columns(self):
        if self.is_scalar_relation:
            # If the attribute is a foreign key, return the Column that this
            # attribute is mapped from -- e.g., .user -> .user_id.
            return _foreign_keys(self._property)
        elif isinstance(self._impl, ScalarAttributeImpl) or self._impl.__class__.__name__ in ('ProxyImpl', '_ProxyImpl'): # 0.4 compatibility: ProxyImpl is a one-off class for each synonym, can't import it
            # normal property, mapped to a single column from the main table
            return self._property.columns
        else:
            # collection -- use the mapped class's PK
            assert self.is_collection, self._impl.__class__
            return self._property.mapper.primary_key
    _columns = property(_columns)

    def relation_type(self):
        """
        The type of object in the collection (e.g., `User`).
        Calling this is only valid when `is_relation` is True.
        """
        return self._property.mapper.class_

    def _pkify(self, value):
        """return the PK for value, if applicable"""
        if value is None:
            return None
        if self.is_collection:
            return [_pk(item) for item in value]
        return value

    def model_value(self):
        return self._pkify(self.raw_value)
    model_value = property(model_value)

    def raw_value(self):
        try:
            v = getattr(self.model, self.name)
        except AttributeError:
            v = getattr(self.model, self.key)
        if v is not None:
            return v

        _columns = self._columns
        if len(_columns) == 1 and  _columns[0].default:
            try:
                from sqlalchemy.sql.expression import Function
            except ImportError:
                from sqlalchemy.sql.expression import _Function as Function
            arg = _columns[0].default.arg
            if callable(arg) or isinstance(arg, Function):
                # callables often depend on the current time, e.g. datetime.now or the equivalent SQL function.
                # these are meant to be the value *at insertion time*, so it's not strictly correct to
                # generate a value at form-edit time.
                pass
            else:
                return arg
        return None
    raw_value = property(raw_value)

    def sync(self):
        """Set the attribute's value in `model` to the value given in `data`"""
        if not self.is_readonly():
            setattr(self.model, self.name, self._deserialize())

    def __eq__(self, other):
        # we override eq so that when we configure with options=[...], we can match the renders in options
        # with the ones that were generated at FieldSet creation time
        try:
            return self._impl is other._impl and _model_equal(self.model, other.model)
        except (AttributeError, ValueError):
            return False
    def __hash__(self):
        return hash(self._impl)

    def __repr__(self):
        return 'AttributeField(%s)' % self.key

    def render(self):
        if self.is_readonly():
            return self.render_readonly()
        if self.is_relation and self.render_opts.get('options') is None:
            if self.is_required() or self.is_collection:
                self.render_opts['options'] = []
            else:
                self.render_opts['options'] = [self._null_option]
            # todo 2.0 this does not handle primaryjoin (/secondaryjoin) alternate join conditions
            fk_cls = self.relation_type()
            order_by = self._property.order_by or list(class_mapper(fk_cls).primary_key)
            q = self.query(fk_cls).order_by(order_by)
            self.render_opts['options'] += _query_options(q)
            logger.debug('options for %s are %s' % (self.name, self.render_opts['options']))
        if self.is_collection and isinstance(self.renderer, self.parent.default_renderers['dropdown']):
            self.render_opts['multiple'] = True
            if 'size' not in self.render_opts:
                self.render_opts['size'] = 5
        return AbstractField.render(self)

    def _get_renderer(self):
        if self.is_relation:
            return self.parent.default_renderers['dropdown']
        return AbstractField._get_renderer(self)

    def _deserialize(self):
        # for multicolumn keys, we turn the string into python via _simple_eval; otherwise,
        # the key is just the raw deserialized value (which is already an int, etc., as necessary)
        if len(self._columns) > 1:
            python_pk = _simple_eval
        else:
            python_pk = lambda st: st

        if self.is_collection:
            return [self.query(self.relation_type()).get(python_pk(pk)) for pk in self.renderer.deserialize()]
        if self.is_composite_foreign_key:
            return self.query(self.relation_type()).get(python_pk(self.renderer.deserialize()))
        return self.renderer.deserialize()



# Copyright (C) 2004,2005 by SICEm S.L. and Imendio AB
#               2006 Johan Dahlin
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

from gazpacho.command import ContainerCommand
from gazpacho.commandmanager import command_manager
from gazpacho.i18n import _
from gazpacho.placeholder import Placeholder
from gazpacho.properties import prop_registry, \
     TransparentProperty, CustomChildProperty, UIntType, \
     CommandSetProperty, PropertySetError
from gazpacho.gadget import Gadget
from gazpacho.widgets.base.base import ContainerAdaptor

class TableAdaptor(ContainerAdaptor):

    def create(self, context, interactive=True):
        """ a GtkTable starts with a default size of 1x1, and setter/getter of
        rows/columns expect the GtkTable to hold this number of placeholders,
        so we should add it. """
        table = super(TableAdaptor, self).create(context, interactive)
        table.attach(Placeholder(), 0, 1, 0, 1)
        return table

    def post_create(self, context, table, interactive=True):
        gadget = Gadget.from_widget(table)
        gadget.get_prop('n-rows').set(3)
        gadget.get_prop('n-columns').set(3)

    def fill_empty(self, context, widget):
        pass

class TableSize(TransparentProperty, UIntType):
    def _add_placeholders(self, table, value, old_size):
        n_columns = table.get_property('n-columns')
        n_rows = table.get_property('n-rows')

        if self.name == 'n-rows':
            table.resize(value, n_columns)
            for col in range(n_columns):
                for row in range(old_size, value):
                    table.attach(Placeholder(), col, col+1, row, row+1)
        else:
            table.resize(n_rows, value)
            for row in range(n_rows):
                for col in range(old_size, value):
                    table.attach(Placeholder(), col, col+1, row, row+1)

    def _remove_children(self, table, value):
        # Remove from the bottom up

        if self.name == 'n-rows':
            start_prop = 'top-attach'
            end_prop = 'bottom-attach'
        else:
            start_prop = 'left-attach'
            end_prop = 'right-attach'

        for child in table.get_children()[::-1]:
            # We need to completely remove it
            start = table.child_get_property(child, start_prop)
            if start >= value:
                table.remove(child)
                continue

            # If the widget spans beyond the new border, we should resize it to
            # fit on the new table
            end = table.child_get_property(child, end_prop)
            if end > value:
                if end_prop == 'right-attach' and value == 0:
                    continue
                table.child_set_property(child, end_prop, value)

        if self.name == 'n-rows':
            table.resize(value, self.get())
        else:
            table.resize(self.get(), value)

    def set(self, value):
        self._value = value
        if value >= 1:
            table = self.object
            old_size = table.get_property(self.name)
            if value > old_size:
                self._add_placeholders(table, value, old_size)
            elif value < old_size:
                self._remove_children(table, value)

    def load(self):
        self._initial = self.object.get_property(self.name)

    def get(self):
        return self._value

    def save(self):
        return str(self.get())

prop_registry.override_simple('GtkTable::n-rows', TableSize)
prop_registry.override_simple('GtkTable::n-columns', TableSize)

prop_registry.override_simple('GtkTable::row-spacing',
                              minimum=0, maximum=10000)
prop_registry.override_simple('GtkTable::column-spacing',
                              minimum=0, maximum=10000)

class BaseAttach(CustomChildProperty):
    """Base class for LeftAttach, RightAttach, TopAttach and BottomAttach
    adaptors"""

    editable = False

    # virtual methods that should be implemented by subclasses:
    def _is_growing(self, value):
        """Returns true if the child widget is growing"""

    def _is_shrinking(self, value):
        """Returns true if the child widget is shrinking"""

    def _cover_placeholder(self, value, left, right, top, bottom):
        """Return True if there is a placeholder in these coordinates"""

    # Private

    def _get_attach(self, table, child):
        """Returns the four attach packing properties in a tuple"""
        right = table.child_get_property(child, 'right-attach')
        left = table.child_get_property(child, 'left-attach')
        top = table.child_get_property(child, 'top-attach')
        bottom = table.child_get_property(child, 'bottom-attach')
        return (left, right, top, bottom)

    def _cell_empty(self, table, x, y):
        """Returns true if the cell at x, y is empty. Exclude child from the
        list of widgets to check"""
        empty = True
        for w in table.get_children():
            left, right, top, bottom = self._get_attach(table, w)
            if (left <= x and (x + 1) <= right and
                top <= y and (y + 1) <= bottom):
                empty = False
                break

        return empty

    # Public

    def set(self, value):
        table = self.object.get_parent()

        # are we growing?
        if self._is_growing(value):
            # check if we need to remove some placeholder
            for ph in [w for w in table.get_children() if isinstance(w, Placeholder)]:
                lph, rph, tph, bph = self._get_attach(table, ph)
                if self._cover_placeholder(table, self.object, value,
                                           lph, rph, tph, bph):
                    table.remove(ph)

            table.child_set_property(self.object, self.name, value)

        # are we shrinking? maybe we need to create placeholders
        elif self._is_shrinking(value):
            table.child_set_property(self.object, self.name, value)
            n_columns = table.get_property('n-columns')
            n_rows = table.get_property('n-rows')
            for y in range(n_rows):
                for x in range(n_columns):
                    if self._cell_empty(table, x, y):
                        table.attach(Placeholder(), x, x+1, y, y+1)

        super(BaseAttach, self).set(value)

class LeftAttach(BaseAttach, UIntType):
    label = 'Left attachment'

    def _is_growing(self, value):
        return value < self.get()

    def _is_shrinking(self, value):
        return value > self.get()

    def _cover_placeholder(self, table, child, value,
                           left, right, top, bottom):
        top_attach = table.child_get_property(child, 'top-attach')
        bottom_attach = table.child_get_property(child, 'bottom-attach')
        if value < right and self.get() > left:
            if top >= top_attach and bottom <= bottom_attach:
                return True
            return False

prop_registry.override_simple_child('GtkTable::left-attach', LeftAttach)

class RightAttach(BaseAttach, UIntType):
    label = 'Right attachment'

    def _is_growing(self, value):
        return value > self.get()

    def _is_shrinking(self, value):
        return value < self.get()

    def _cover_placeholder(self, table, child, value,
                           left, right, top, bottom):
        top_attach = table.child_get_property(child, 'top-attach')
        bottom_attach = table.child_get_property(child, 'bottom-attach')
        if value > left and self.get() < right:
            if top >= top_attach and bottom <= bottom_attach:
                return True
        return False

prop_registry.override_simple_child('GtkTable::right-attach', RightAttach)

class BottomAttach(BaseAttach, UIntType):
    label = 'Bottom attachment'

    def _is_growing(self, value):
        return value > self.get()

    def _is_shrinking(self, value):
        return value < self.get()

    def _cover_placeholder(self, table, child,
                           value, left, right, top, bottom):
        right_attach = table.child_get_property(child, 'right-attach')
        left_attach = table.child_get_property(child, 'left-attach')
        if value > top and self.get() < bottom:
            if left >= left_attach and right <= right_attach:
                return True
        return False

prop_registry.override_simple_child('GtkTable::bottom-attach', BottomAttach)

class TopAttach(BaseAttach, UIntType):
    label = 'Top attachment'

    def _is_growing(self, value):
        return value < self.get()

    def _is_shrinking(self, value):
        return value > self.get()

    def _cover_placeholder(self, table, child,
                           value, left, right, top, bottom):
        right_attach = table.child_get_property(child, 'right-attach')
        left_attach = table.child_get_property(child, 'left-attach')
        if value < bottom and self.get() > top:
            if left >= left_attach and right <= right_attach:
                return True
        return False

prop_registry.override_simple_child('GtkTable::top-attach', TopAttach)

class CommonCustom(CustomChildProperty, UIntType):
    def __init__(self, gadget):
        super(CustomChildProperty, self).__init__(gadget)
        self._update_lock = False

    def get(self):
        return self._value

    def save(self):
        return

    def _notify_update(self, gobj, pspec):
        parent = gobj.get_parent()
        self._update_lock = True
        self.set(parent.child_get_property(gobj, pspec.name))
        self._update_lock = False

    def set(self, value):
        if not self._update_lock:
            self._update_value(value)

        self._value = value
        self.notify()

class XPos(CommonCustom):
    label = 'X position'
    priority = 1

    def load(self):
        object = self.object
        parent = self.object.get_parent()
        self._value = parent.child_get_property(object, 'left-attach')
        self.object.connect('child-notify::left-attach', self._notify_update)

    def _update_value(self, value):
        gadget = self.gadget
        n_columns = gadget.get_parent().get_prop('n-columns').value
        left = gadget.get_child_prop('left-attach')
        right = gadget.get_child_prop('right-attach')
        span = right.value - left.value

        if value > n_columns - span:
            raise PropertySetError

        diff = value - self._value

        # Order is important, GtkTable does not allow us to have
        # a child which does not take up any space at all, eg
        # left-attach = right-attach
        cmd_left = CommandSetProperty(left, left.value + diff)
        cmd_right = CommandSetProperty(right, right.value + diff)
        if diff > 0:
            cmds = [cmd_right, cmd_left]
        else:
            cmds = [cmd_left, cmd_right]
        command_manager.execute(ContainerCommand(
            _('Moving widget to x position %d') % value, *cmds),
                                self.gadget.project, nested=False)

prop_registry.override_child_property('GtkTable::x-pos', XPos)

class YPos(CommonCustom):
    label = 'Y position'
    priority = 2

    def load(self):
        object = self.object
        parent = self.object.get_parent()
        self._value = parent.child_get_property(object, 'top-attach')
        self.object.connect('child-notify::top-attach', self._notify_update)

    def _update_value(self, value):
        gadget = self.gadget
        n_rows = gadget.get_parent().get_prop('n-rows').value
        top = gadget.get_child_prop('top-attach')
        bottom = gadget.get_child_prop('bottom-attach')
        span = bottom.value - top.value

        if value > n_rows - span:
            raise PropertySetError

        diff = value - self._value

        # Order is important, GtkTable does not allow us to have
        # a child which does not take up any space at all, eg
        # top-attach = bottom-attach
        cmd_bottom = CommandSetProperty(bottom, bottom.value + diff)
        cmd_top = CommandSetProperty(top, top.value + diff)
        if diff > 0:
            cmds = [cmd_bottom, cmd_top]
        else:
            cmds = [cmd_top, cmd_bottom]
        command_manager.execute(ContainerCommand(
            _('Moving widget to y position %d') % value, *cmds),
                                self.gadget.project, nested=False)

prop_registry.override_child_property('GtkTable::y-pos', YPos)

class ColSpan(CommonCustom):
    label = 'Colspan'
    minimum = 1
    priority = 3

    def _notify_update(self, gobj, pspec):
        parent = gobj.get_parent()
        self._update_lock = True
        self.set((parent.child_get_property(gobj, 'right-attach') -
                  parent.child_get_property(gobj, 'left-attach')))
        self._update_lock = False

    def load(self):
        object = self.object
        parent = self.object.get_parent()

        self._value = (parent.child_get_property(object, 'right-attach') -
                       parent.child_get_property(object, 'left-attach'))
        self.object.connect('child-notify::left-attach', self._notify_update)
        self.object.connect('child-notify::right-attach', self._notify_update)

    def _update_value(self, value):
        gadget = self.gadget
        n_columns = gadget.get_parent().get_prop('n-columns').value
        x_pos = gadget.get_child_prop('left-attach').value

        if value > n_columns - x_pos:
            raise PropertySetError

        right = gadget.get_child_prop('right-attach')
        command_manager.execute(
            CommandSetProperty(right, right.value + (value - self._value)),
            self.gadget.project,
            nested=False)

prop_registry.override_child_property('GtkTable::colspan', ColSpan)

class RowSpan(CommonCustom):
    label = 'Rowspan'
    minimum = 1
    priority = 4

    def _notify_update(self, gobj, pspec):
        parent = gobj.get_parent()
        self._update_lock = True
        self.set((parent.child_get_property(gobj, 'bottom-attach') -
                  parent.child_get_property(gobj, 'top-attach')))
        self._update_lock = False

    def load(self):
        object = self.object
        parent = self.object.get_parent()
        self._value = (parent.child_get_property(object, 'bottom-attach') -
                       parent.child_get_property(object, 'top-attach'))
        self.object.connect('child-notify::bottom-attach',
                            self._notify_update)
        self.object.connect('child-notify::top-attach',
                            self._notify_update)

    def _update_value(self, value):
        gadget = self.gadget
        n_rows = gadget.get_parent().get_prop('n-rows').value
        y_pos = gadget.get_child_prop('top-attach').value

        if value > n_rows - y_pos:
            raise PropertySetError

        bottom = gadget.get_child_prop('bottom-attach')
        command_manager.execute(
            CommandSetProperty(bottom, bottom.value + (value - self._value)),
            self.gadget.project,
            nested=False)

prop_registry.override_child_property('GtkTable::rowspan', RowSpan)



# Copyright (C) 2004,2005 by SICEm S.L.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
import gobject
import gtk
from kiwi.environ import environ
from kiwi.utils import gsignal

from gazpacho.catalog import get_all_catalogs, get_custom_ui_filename, CustomUI
from gazpacho.dndhandlers import DND_PALETTE_TARGET, INFO_TYPE_PALETTE
from gazpacho.i18n import _

class Palette(gtk.VBox):
    gsignal('toggled')

    def __init__(self):
        """Create a palette widget for a list of catalogs.

        A Palette widget graphically represents a set of widget groups
        as described in catalog files and allows for the notification
        of selection changes and the currently selected element.

        A Palette has the following elements: a selector, a button group and
        a notebook.

        The selector allows to toggle between "select widget" mode and
        "add widget" mode.

        The button group has a radio button for each group in the catalogs,
        which additionally activates the group's own notebook page.

        The notebook has a page for each existing catalog group. These
        pages show a row per widget and each row holds an icon and a label
        to represent the widget element. This data is taken from the catalog.

        Derived classes connect to the Palette's "toggled" signal
        to be notified of selection changes so they may read the
        "current" property holding the selected "widget_adaptor" and react
        accordingly.
        """
        gtk.VBox.__init__(self)

        # The WidgetAdaptor corresponding to the selected button. None if
        # the selector button is pressed.
        self._current = None

        self.pack_start(self._selector_new(), False)
        self.pack_start(gtk.HSeparator(), False, padding=3)

        #The vbox that contains the titles of the sections in buttons
        self._groups_vbox = gtk.VBox()
        self.pack_start(self._groups_vbox, False)
        self.pack_start(gtk.HSeparator(), False, padding=3)

        #  Where we store the different catalogs
        self._notebook = gtk.Notebook()
        self._notebook.set_show_tabs(False)
        self._notebook.set_show_border(False)
        self.pack_end(self._notebook)

        # The number of sections in this palette
        self._nb_sections = 0

        self._tooltips = gtk.Tooltips()

        # Each section in the palette has a radio button. This is the
        # button group to which new section buttons must be appended
        self._sections_button_group = None

        catalogs = get_all_catalogs()
        custom_ui_filename = get_custom_ui_filename()
        if custom_ui_filename:
            custom_ui = CustomUI(custom_ui_filename)
            for group in custom_ui.widget_groups:
                self.append_widget_group(group)
        else:
            for catalog in catalogs:
                for group in catalog.widget_groups:
                    self.append_widget_group(group)

        # persistent mode means that a widget keep selected after
        # creating an instance of it. It is activated when the
        # user holds SHIFT while selecting the widget
        self.persistent_mode = False

    # some needed property accesors for the currently selected element
    def get_current(self):
        return self._current

    def set_current(self, value):
        self._current = value
    current = property(get_current, set_current)

    def _selector_new(self):
        """Return a new selector.

        The selector is a button used to switch between "widget selection"
        and "widget add" mode. This makes the cursor look change. When no
        widget is selected in the notebook then the selector button is pressed.
        """
        hbox = gtk.HBox()
        # The selector is part of the widgets_button_group, so that when
        # no widget is selected, its button is pressed.
        self._selector = gtk.RadioButton(None)
        self._selector.set_name('Selector')
        self._selector.set_mode(False)
        self._selector.set_relief(gtk.RELIEF_NONE)

        # Each widget in a section has a button in the palette. This is the
        # button group, since only one may be pressed.
        self._widgets_button_group = self._selector.get_group()

        image = gtk.Image()
        image.set_from_file(environ.find_resource('pixmap', 'selector.png'))
        self._selector.add(image)
        self._selector.connect('toggled', self._on_button_toggled)

        # A label which contains the name of the class currently selected or
        # "Selector" if no widget class is selected
        self._label = gtk.Label(_('Selector'))
        self._label.set_alignment(0.0, 0.5)

        hbox.pack_start(self._selector, False, False)
        hbox.pack_start(self._label, padding=2)
        hbox.show_all()

        return hbox

    def _on_button_toggled(self, button):
        """Show the selected button label by the selector and make it current.

        Derived classes use the toggled signal to get notified of changes and
        read the "current" property to get the widget_adaptor.
        """
        if not button.get_active():
            return

        self.persistent_mode = False

        if button == self._selector:
            self._current = None
            self._label.set_text(_('Selector'))
        else:
            modifiers = button.window.get_pointer()[2]
            if modifiers & gtk.gdk.CONTROL_MASK:
                self.persistent_mode = True

            self._current = button.get_data('user')
            self._label.set_text(self._current.editor_name)

        self.emit('toggled')

    def _on_catalog_button_toggled(self, button):
        "Show notebook page associated to a catalog button"
        page = button.get_data('page')
        self._notebook.set_current_page(page)
        return True

    def do_toggled(self):
        pass

    def append_widget_group(self, group):
        """Append a new group of widgets to the palette.

        For the given catalog group a new RadioButton is added to control
        the display of the notebook page that holds the widget table. This
        table is also built from this method.
        """
        page = self._nb_sections
        self._nb_sections += 1

        # add the button to the existing group or create a new group
        # to be used from now on.
        if not self._sections_button_group:
            button = gtk.RadioButton(None, group.title)
        else:
            button = gtk.RadioButton(self._sections_button_group,
                                     group.title)
	# FIXME: This needs to be able to handle space, ampersand etc.
        button.set_name(group.name.replace('-', '_'))
        # use 1st button in the group as reference for the group
        # and store the notebook page controlled by the button
        self._sections_button_group = button.get_group()[0]
        button.set_mode(False)
        button.set_data('page', page)
        button.connect('toggled', self._on_catalog_button_toggled)

        self._groups_vbox.pack_start(button, False)

        # add the selection
        self._notebook.append_page(self._widget_table_create(group),
                                   gtk.Label(''))
        self._notebook.show()

    def _setup_dnd(self, widget):
        targets = [DND_PALETTE_TARGET]
        widget.drag_source_set(gtk.gdk.BUTTON1_MASK, targets,
                               gtk.gdk.ACTION_COPY | gtk.gdk.ACTION_MOVE)

        widget.connect('drag_begin', self._dnd_drag_begin_cb)
        widget.connect('drag_data_get', self._dnd_drag_data_get_cb)

    def _widget_table_create(self, group):
        """Return a table with an icon and a label for every widget in a group.

        Each button stores with it a 'user' attribute that holds the
        widget_adaptor denomination from the catalog data.
        """

        vbox = gtk.VBox()
        for widget_adaptor in group:
            if not widget_adaptor.palette_name:
                continue

            radio = gtk.RadioButton(self._widgets_button_group[0])
            radio.set_name(widget_adaptor.generic_name)
            radio.connect('toggled', self._on_button_toggled)
            radio.set_data('user', widget_adaptor)
            radio.set_mode(False)
            radio.set_relief(gtk.RELIEF_NONE)

            # Drag and Drop, but only for non-toplevel widgets
            if not widget_adaptor.is_toplevel():
                self._setup_dnd(radio)
            vbox.pack_start(radio, False, False)

            if widget_adaptor.tooltip:
                self._tooltips.set_tip(radio, widget_adaptor.tooltip)

            hbox = gtk.HBox(spacing=2)
            hbox.pack_start(widget_adaptor.icon, False, False)
            radio.add(hbox)

            label = gtk.Label(widget_adaptor.palette_name)
            label.set_alignment(0.0, 0.5)
            hbox.pack_start(label, padding=1)

            self._widgets_button_group = radio.get_group()

        scrolled_window = gtk.ScrolledWindow()
        scrolled_window.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)
        scrolled_window.add_with_viewport(vbox)
        scrolled_window.set_shadow_type(gtk.SHADOW_NONE)

        return scrolled_window

    def _dnd_drag_begin_cb(self, source_gadget, drag_context):
        """Callback for handling the 'drag_begin' event. It will just
        set a nice looking drag icon."""

        source_gadget.clicked()

        adaptor = source_gadget.get_data('user')

        if adaptor:
            source_gadget.drag_source_set_icon_pixbuf(adaptor.pixbuf)

    def _dnd_drag_data_get_cb(self, source_gadget, context,
                              selection_data, info, time):
        adaptor = source_gadget.get_data('user')

        if adaptor and info == INFO_TYPE_PALETTE:
            selection_data.set(selection_data.target, 8, adaptor.type_name)
            return

        selection_data.set(selection_data.target, 8, "")

    def unselect_widget(self):
        "Leave \"add widget\" mode and back to the \"select widget\" mode."
        self._selector.set_active(True)

    def select_widget(self, catalog, widgetname):
        """Select a widget in the palette from a catalog with name widgetname.

        This selection triggers the clicked signal for the widget button
        in the palette.
        """
        # select the catalog
        pagenum = -1
        for button in self._groups_vbox.get_children():
            if button.get_label() == catalog:
                pagenum = button.get_data('page')
                break

        if pagenum == -1:
            return

        vbox = self._notebook.get_nth_page(pagenum).get_child().get_child()
        # then select the widget
        for button in vbox.get_children():
            wc = button.get_data('user')
            if wc.name == widgetname:
                button.clicked()

gobject.type_register(Palette)

if environ.epydoc:
    palette = object()
else:
    palette = Palette()



#
#    Autocomplete feature for admin panel
#
#    Most of the code has been written by Jannis Leidel:
#    http://jannisleidel.com/2008/11/autocomplete-form-widget-foreignkey-model-fields/
#
#    to_string_function, Satchmo adaptation and some comments added by emes (Michal Salaban)
#
from django import forms
from django.db import models
from django.db.models.query import QuerySet
from django.conf import settings
from django.contrib import admin
from django.contrib.auth.models import Message
from django.http import HttpResponse, HttpResponseNotFound
from django.utils.encoding import smart_str
from django.utils.html import escape
from django.utils.safestring import mark_safe
from django.utils.text import truncate_words

import operator

class ForeignKeySearchInput(forms.HiddenInput):
    """
    A Widget for displaying ForeignKeys in an autocomplete search input 
    instead in a <select> box.
    """

    to_string_function = lambda s: truncate_words(s, 14)

    class Media:
        css = {
            'all': ('css/jquery.autocomplete.css',)
        }
        js = (
            # The js/jquery.js script is referenced in admin/base_site.html template.
            # Requesting it here again would reset all the plugins loaded afterwards.
            'js/jquery.bgiframe.js',
            'js/jquery.ajaxQueue.js',
            'js/jquery.autocomplete.js'
        )

    def label_for_value(self, value):
        key = self.rel.get_related_field().name
        obj = self.rel.to._default_manager.get(**{key: value})
        return self.to_string_function(obj)

    def __init__(self, rel, search_fields, to_string_function, attrs=None):
        self.rel = rel
        self.search_fields = search_fields
        if to_string_function: self.to_string_function = to_string_function
        super(ForeignKeySearchInput, self).__init__(attrs)

    def render(self, name, value, attrs=None):
        if attrs is None:
            attrs = {}
        rendered = super(ForeignKeySearchInput, self).render(name, value, attrs)
        if value:
            label = self.label_for_value(value)
        else:
            label = u''
        return rendered + mark_safe(u'''
            <style type="text/css" media="screen">
                #lookup_%(name)s {
                    padding-right:16px;
                    background: url(
                        %(admin_media_prefix)simg/admin/selector-search.gif
                    ) no-repeat right;
                }
                #del_%(name)s {
                    display: none;
                }
            </style>
<input type="text" id="lookup_%(name)s" value="%(label)s" />
<a href="#" id="del_%(name)s">
<img src="%(admin_media_prefix)simg/admin/icon_deletelink.gif" />
</a>
<script type="text/javascript">
            var lookup = $('#lookup_%(name)s')
            if (lookup.val()) {
                $('#del_%(name)s').show()
            }
            lookup.attr('size', Math.max(10, lookup.attr('value').length))
            lookup.autocomplete('../search/', {
                formatResult: function(data){ return $('<div />').html(data[0]).text(); },
                extraParams: {
                    search_fields: '%(search_fields)s',
                    app_label: '%(app_label)s',
                    model_name: '%(model_name)s'
                }
            }).result(function(event, data, formatted) {
                if (data) {
                    $('#id_%(name)s').val(data[1]);
                    $('#del_%(name)s').show();
                }
            });
            $('#del_%(name)s').click(function(ele, event) {
                $('#id_%(name)s').val('');
                $('#del_%(name)s').hide();
                $('#lookup_%(name)s').val('');
            });
            </script>
        ''') % {
            'search_fields': ','.join(self.search_fields),
            'admin_media_prefix': settings.ADMIN_MEDIA_PREFIX,
            'model_name': self.rel.to._meta.module_name,
            'app_label': self.rel.to._meta.app_label,
            'label': label,
            'name': name,
        }

class AutocompleteAdmin(admin.ModelAdmin):
    """Admin class for models using the autocomplete feature.

    There are two additional fields:
       - related_search_fields: defines fields of managed model that
         have to be represented by autocomplete input, together with
         a list of target model fields that have to be searched for
         input string,
       - related_string_functions: contains optional functions which
         take target model instance as only argument and return string
         representation. By default __unicode__() method of target
         object is used.
    """

    related_search_fields = {}
    related_string_functions = {}

    def __call__(self, request, url):
        # This is deprecated interface and will be dropped in Django 1.3.
        # Since the version 1.1, Django uses get_urls() method below.
        if url is None:
            pass
        elif url == 'search':
            return self.search(request)
        return super(AutocompleteAdmin, self).__call__(request, url)

    def get_urls(self):
        from django.conf.urls.defaults import url
        patterns = super(AutocompleteAdmin, self).get_urls()
        info = self.admin_site.name, self.model._meta.app_label, self.model._meta.module_name
        patterns.insert(
                -1,     # insert just before (.+) rule (see django.contrib.admin.options.ModelAdmin.get_urls)
                url(
                    r'^search/$',
                    self.search,
                    name='%sadmin_%s_%s_search' % info
                    )
                )
        return patterns

    def search(self, request):
        """
        Searches in the fields of the given related model and returns the
        result as a simple string to be used by the jQuery Autocomplete plugin
        """
        query = request.GET.get('q', None)
        app_label = request.GET.get('app_label', None)
        model_name = request.GET.get('model_name', None)
        search_fields = request.GET.get('search_fields', None)
        try:
            to_string_function = self.related_string_functions[model_name]
        except KeyError:
            to_string_function = lambda x: x.__unicode__()

        if search_fields and app_label and model_name and query:
            def construct_search(field_name):
                # use different lookup methods depending on the notation
                if field_name.startswith('^'):
                    return "%s__istartswith" % field_name[1:]
                elif field_name.startswith('='):
                    return "%s__iexact" % field_name[1:]
                elif field_name.startswith('@'):
                    return "%s__search" % field_name[1:]
                else:
                    return "%s__icontains" % field_name

            model = models.get_model(app_label, model_name)
            qs = model._default_manager.all()
            for bit in query.split():
                or_queries = [models.Q(**{construct_search(
                    smart_str(field_name)): smart_str(bit)})
                        for field_name in search_fields.split(',')]
                other_qs = QuerySet(model)
                other_qs.dup_select_related(qs)
                other_qs = other_qs.filter(reduce(operator.or_, or_queries))
                qs = qs & other_qs
            data = ''.join([u'%s|%s\n' % (to_string_function(f), f.pk) for f in qs])
            return HttpResponse(data)
        return HttpResponseNotFound()

    def formfield_for_dbfield(self, db_field, **kwargs):
        """
        Overrides the default widget for Foreignkey fields if they are
        specified in the related_search_fields class attribute.
        """
        if isinstance(db_field, models.ForeignKey) and \
                db_field.name in self.related_search_fields:
            kwargs['widget'] = ForeignKeySearchInput(
                    db_field.rel,
                    self.related_search_fields[db_field.name],
                    self.related_string_functions.get(db_field.name)
                    )
        field = super(AutocompleteAdmin, self).formfield_for_dbfield(db_field, **kwargs)
        return field



#! /usr/bin/env python

from optparse import OptionParser
import os.path
import sys

__version__ = "0.1"

USAGE = """pybuild [options] filename1 filename2 ..."""

def posplit(fname, rules):
    sections = make_sections(fname)
    rules.set_header(sections[0])
    rules.set_pofile(fname)
    print "SECTIONS: %s" % fname
    ix = 0
    for section in sections[1:]:
        section.apply_rules(rules)
        section.do_copy()
        ix += 1
        #print ix
    
def make_sections(fname):
    sections = []
    f = open(fname, 'r')
    section=Section()
    for line in f.xreadlines():
        if line and line.endswith("\n"):
            line = line[:-1]
        elif line and (line.endswith("\r\n") or line.endswith("\n\r")):
            line = line[:-2]
            
        if not line:
            sections.append(section)
            section = Section()
        else:
            section.add(line)
            
    return sections

class RuleSet(object):
    
    def __init__(self, fname):
        self.rules = []
        self.fname = fname
        self.parse_rules()
        self.files = {}
        
    def parse_rules(self):
        """Parse a rules file which tells where to put target sections

        The rules file is formatted like:
        slug=[option:]dest

        Where:
        - slug is the beginning part of the filename in the po file.
        - dest is the destination name

        For example:
        accounts=apps/satchmo_store/accounts
        giftcertificate=apps/payment/modules/giftcertificate
        local_settings-customize.py=nostrip:projects/template/
        """
        f = open(self.fname, 'r')
        for line in f.xreadlines():
            if line and line.endswith("\n"):
                line = line[:-1]
            if line and not line.startswith("#"):
                self.add(Rule(line))

    def add(self, rule):
        rule.parent = self
        self.rules.append(rule)
    
    def match_rule(self, target):
        decision = None
        for rule in self.rules:
            if rule.applies_to(target):
                decision = rule
                break
        
        return decision
        
    def open_pofile(self, dest):
        if not dest in self.files:
            #print "making: %s" % dest
            fname = os.path.join(dest, self.pofile)
            d, n = os.path.split(fname)
            if not os.path.exists(d):
                print "Making directory: %s" % d
                os.makedirs(d)
            print "opening: %s" % fname
            f = open(fname, 'w')
            f.write("\n".join(self.header.lines))
            f.write("\n\n")
            self.files[dest] = f
        else:
            pass
            #print "reusing: %s" % dest
        
        return self.files[dest]

    def set_header(self, header):
        self.header = header
        
    def set_pofile(self, fname):
        self.pofile = fname
        
class Rule(object):
    def __init__(self, line):
        try:
            k, v = line.split('=')
        except:
            print "error: '%s'" % line
            raise
        opt = "STRIP"
        if v.find(':') > -1:
            opt, v = v.split(':')
                
        self.parent = None
        self.key = k
        self.option = opt
        self.dest = v

    def applies_to(self, target):
        return target.startswith(self.key)

    def apply(self, target):
        if self.option == "IGNORE":
            return ""
            
        if self.option == "STRIP":
            if target.startswith(self.key):
                target = target[len(self.key):]
        
        return target
        
    def do_copy(self, targets, lines):
        if not self.option=='IGNORE':
            f = self.parent.open_pofile(self.dest)
            targets = ["#: %s" % self.apply(tgt) for tgt in targets]
            f.write("\n".join(targets))
            f.write("\n")
            f.write("\n".join(lines))
            f.write("\n\n")

    def __str__(self):
        return("Rule: %s=(%s) %s" % (self.key, self.option, self.dest))

class Section(object):
    
    def __init__(self):
        self.targets = []
        self.decisions = {}
        self.lines = []
        
    def add(self, line):
        if line.startswith("#: "):
            self.targets.append(line[3:])
        else:
            self.lines.append(line)
            
    def apply_rules(self, rules):
        for target in self.targets:
            decision = rules.match_rule(target)
            
            if decision:
                if decision in self.decisions:
                    self.decisions[decision].append(target)
                else:
                    self.decisions[decision] = [target]
            else:
                print "WARNING, no decision made for target: '%s'" % target
                sys.exit(3)
            
    def do_copy(self):
        for rule, targets in self.decisions.items():
            rule.do_copy(targets, self.lines)

    def __unicode__(self):
        out = ['TARGETS: ',]
        out.extend(self.targets)
        out.append('DECISIONS: ')
        for k, v in self.decisions.items():
            out.append('%s = %s' % (k.key, ", ".join(v)))

        out.append('LINES:')
        out.extend(self.lines)
        return "\n".join(out)


def main(args):
    parser = OptionParser(USAGE, version = __version__)
    
    parser.add_option('-c','--conf', dest='conf',
                      default='posplit.rules', 
                      help="the posplit rules file")
    
    opts, fnames = parser.parse_args(args)
    
    if not fnames:
        print "Need at least one filename"
        print USAGE
        sys.exit(2)
        
    if not opts.conf:
        print "Need a config file"
        print USAGE
        sys.exit(2)
        
    rules = RuleSet(opts.conf)
    
    for fname in fnames:
        posplit(fname, rules)
    
if __name__=='__main__':
    main(sys.argv[1:])


import sys
#import mainapp
from notmm.utils.wsgilib import HTTPRequest
from test_support import (
    WSGIControllerTestCase, 
    TestClient, 
    unittest,
    make_app,
    settings,
    ResponseClass
    )

class WSGIAppTestCase(WSGIControllerTestCase):

    def test_with_session(self):
        response = self.client.get('/test_with_session') # 
        self.assertEqual(response.status_code, '200 OK')




#from sandbox_sqlalchemy.config.settings import *
#from sandbox.config.settings import *
from moviereviews.config.settings import *

DEBUG = True
ENABLE_BEAKER = True
SECRET_KEY = '12344dsdsd'
TEMPLATE_LOADER = 'notmm.utils.template.backends.CachedTemplateLoader'

PUBSUB_ERROR_HANDLERS = CUSTOM_ERROR_HANDLERS



"""urlpatterns for productratings.  Note that you do not need to add these to your urls anywhere, they'll be automatically added by the collect_urls signals."""

from django.conf.urls.defaults import *
import logging

log = logging.getLogger('productratings.urls')

productpatterns = patterns('satchmo_ext.productratings.views',
    (r'^view/bestrated/$', 
        'display_bestratings', {}, 'satchmo_product_best_rated'),
)

# Override comments with our redirecting view. You can remove the next two
# URLs if you aren't using ratings.
#(r'^comments/post/$', 'comments.post_rating', {'maxcomments': 1 }, 'satchmo_rating_post'),
commentpatterns = patterns('',
    (r'^comments/', include('django.contrib.comments.urls')),
)

def add_product_urls(sender, patterns=(), section="", **kwargs):
    if section=="product":
        log.debug('adding ratings urls')
        patterns += productpatterns

def add_comment_urls(sender, patterns=(), **kwargs):
    log.debug('adding comments urls')
    patterns += commentpatterns



# Suck in default django settings
from django.conf.global_settings import *
from pkg_resources import resource_filename
import os

DEBUG = True

USE_I18N = False

# Custom error handlers mapping
CUSTOM_ERROR_HANDLERS = (
    ('handle401', 'moviereviews.handlers.handle401'),
    #('handle403', 'moviereviews.handlers.handle403'),
    ('handle404', 'moviereviews.handlers.handle404'),
    ('handle500', 'moviereviews.handlers.handle500')
    )
PUBSUB_ERROR_HANDLERS = CUSTOM_ERROR_HANDLERS

# List of processors used by RequestContext to populate the context.
# Each one should be a callable that takes the request object as its
# only parameter and returns a dictionary to add to the context.
TEMPLATE_CONTEXT_PROCESSORS = (
    'django.contrib.auth.context_processors.auth',
    )

# For backward-compatibility with Django
TEMPLATE_DIRS = (
    resource_filename('moviereviews', 'templates'),
    resource_filename('blogengine', 'templates')
)

ROOT_URLCONF = 'moviereviews.config.urls'
MIDDLEWARE_CLASSES = ()

# XXX deprecated
# DATABASE_NAME = 'sqlite:///' + resource_filename('moviereviews_generic', 'fixtures/test.db')
DATABASE_NAME = 'moviereviews'

SITE_ID = 1

INSTALLED_APPS = (
    'moviereviews',
)

SCHEVO = {
    'DATABASE_NAME': 'moviereviews',
    'DATABASE_PREFIX': 'schevo.db',
    'ENABLE_GZIP': True,
    'DEBUG' : True
    }




"""Allows livesettings to be "locked down" and no longer use the settings page or the database
for settings retrieval.
"""

from django.conf import settings as djangosettings
from django.contrib.sites.models import Site
from django.db import transaction

__all__ = ['get_overrides']

def _safe_get_siteid(site):
    if not site:
        try:
            site = Site.objects.get_current()
        except:
            transaction.rollback()
        if site and site.id:
            siteid = site.id
        else:
            siteid = djangosettings.SITE_ID
    else:
        siteid = site.id
    transaction.commit()
    return siteid

#_safe_get_siteid=transaction.commit_manually(_safe_get_siteid)

def get_overrides(siteid=-1):
    """Check to see if livesettings is allowed to use the database.  If not, then
    it will only use the values in the dictionary, LIVESETTINGS_OPTIONS[SITEID]['SETTINGS'],
    this allows 'lockdown' of a live site.

    The LIVESETTINGS dict must be formatted as follows::

    LIVESETTINGS_OPTIONS = {
            1 : {
                    'DB' : [True/False],
                    SETTINGS = {
                        'GROUPKEY' : {'KEY', val, 'KEY2', val},
                        'GROUPKEY2' : {'KEY', val, 'KEY2', val},
                    }
                }
            }

    In the settings dict above, the "val" entries must exactly match the format 
    stored in the database for a setting.  Do not use a literal True or an integer,
    it needs to be the string representation of them.

    Returns a tuple (DB_ALLOWED, SETTINGS)
    """
    if siteid == -1:
        siteid = _safe_get_siteid(None)

    overrides = (True, {})
    if hasattr(djangosettings, 'LIVESETTINGS_OPTIONS'):
        opts = djangosettings.LIVESETTINGS_OPTIONS
        if opts.has_key(siteid):
            opts = opts[siteid]
            overrides = (opts['DB'], opts['SETTINGS'])

    return overrides



# coding=UTF-8
from django.db import models
from django.utils.translation import ugettext, ugettext_lazy as _
from product.models import TaxClass
from l10n.models import AdminArea, Country
#from satchmo_store.shop.models import Order
#from satchmo_store.shop.signals import order_success
#from tax import Processor

from datetime import date as _date

try:
    from decimal import Decimal
except:
    from django.utils._decimal import Decimal

class Taxable(models.Model):
    """
    Map that says what items are taxable in a jurisdiction.

    To use properly, assign products to a meaningful TaxClass, such as 'Shipping',
    'Food', 'Default'. Then create rules for the jurisdictions where you are
    required to collect tax. If for example, you are taxing objects in two states
    and 'Food' is taxable in one and not the other, but shipping is the other
    way around, you would need to create the following entries:

    food = TaxClass(...)
    default = TaxClass(...)
    shipping = TaxClass(...)

    one_state = AdminArea(...)
    two_state = AdminArea(...)
    usa = Country(...)

    Taxable(taxClass=default,  isTaxable=True,  taxZone=one_state, taxCountry=usa)
    Taxable(taxClass=food,     isTaxable=False, useFood=True, taxZone=one_state, taxCountry=usa)
    Taxable(taxClass=shipping, isTaxable=True,  taxZone=one_state, taxCountry=usa)
    Taxable(taxClass=default,  isTaxable=True,  taxZone=two_state, taxCountry=usa)
    Taxable(taxClass=food,     isTaxable=True,  useFood=True, taxZone=two_state, taxCountry=usa)
    Taxable(taxClass=shipping, isTaxable=False, taxZone=two_state, taxCountry=usa)

    Laws vary drastically form state to state, so please make sure to make needed
    TaxClasses for all objects that vary in taxing jurisdictions to which you
    must submit.

    If you do not at least create a 'Default' entry for a state, then you will
    not be collecting any taxes for that state. Only create entires for states
    where you are obligated to collect and report taxes.

    SST defines food rates and interstate vs. intrastate rates. You may override
    these, otherwise taxes will be charged at the non-food, intrastate rate by default.

    WARNING: If a product is taxable in ANY jurisdiction, it must be set taxable
    in the product. You disable it per-jurisdiction by disabling it here. You
    cannot enable it here if it is disabled on the product itself.
    """
    taxClass = models.ForeignKey(TaxClass, verbose_name=_('Tax Class'))
    taxZone = models.ForeignKey(AdminArea, blank=True, null=True,
        verbose_name=_('Tax Zone'))
    taxCountry = models.ForeignKey(Country, blank=True, null=True,
        verbose_name=_('Tax Country'))
    isTaxable = models.BooleanField(verbose_name=_('Taxable?'), default=True, )
    useIntrastate = models.BooleanField(verbose_name=_('Use Intrastate rate instead of Interstate?'),
                                        default=True)
    useFood = models.BooleanField(verbose_name=_('Use food/drug rate instead of general?'),
                                  default=False)

    def _country(self):
        if self.taxZone:
            return self.taxZone.country.name
        else:
            return self.taxCountry.name
    country = property(_country)

    #def _display_percentage(self):
    #    return "%#2.2f%%" % (100*self.percentage)
    #_display_percentage.short_description = _('Percentage')
    #display_percentage = property(_display_percentage)

    def __unicode__(self):
        return u"%s - %s = %s" % (self.taxClass,
                             self.taxZone and self.taxZone or self.taxCountry,
                             self.isTaxable)

    class Meta:
        verbose_name = _("Taxable Class")
        verbose_name_plural = _("Taxable Classes")

JURISDICTION_CHOICES = (
    (0, 'County'),
    (1, 'City'),
    (2, 'Town'),
    (3, 'Village'),
    (4, 'Borough'),
    (5, 'Township'),
    (9, 'Other Municipality'),
    (10, 'School District'),
    (11, 'Junior Colleges'),
    (19, 'Other Schools'),
    (20, 'Water Control'),
    (21, 'Utility District'),
    (22, 'Sanitation'),
    (23, 'Water or Sewer District'),
    (24, 'Reclamation District'),
    (25, 'Fire or Police'),
    (26, 'Roads or Bridges'),
    (27, 'Hospitals'),
    (29, 'Other Municipal Services'),
    (40, 'Township and County'),
    (41, 'City and School'),
    (42, 'County collected by Other Taxing Authority'),
    (43, 'State and County'),
    (44, 'Central Collection Taxing Authority'),
    (45, 'State Taxing Authority'),
    (49, 'Other Combination Collection'),
    (50, 'Bond Authority'),
    (51, 'Annual County Bond Authority'),
    (52, 'Semi-annual County Bond Authority'),
    (53, 'Annual City Bond Authority'),
    (54, 'Semi-annual City Bond Authority'),
    (59, 'Other Bond Authority'),
    (61, 'Assessment District'),
    (62, 'Homeowner’s Association'),
    (63, 'Special District'),
    (69, 'Other Special Districts'),
    (70, 'Central Appraisal Taxing Authority'),
    (71, 'Unsecured County Taxes'),
    (72, 'Mobile Home Authority'),
    (79, 'Other Special Applications'),
)

class TaxRate(models.Model):
    """
    Records for tax rates in the default SST format as defined at:
    http://www.streamlinedsalestax.org/Technology/RatesandBoundariesClean082605.pdf
    """
    state = models.IntegerField(max_length=2,
        verbose_name=_('FIPS State Code'), db_index=True)
    jurisdictionType = models.IntegerField(max_length=2, choices=JURISDICTION_CHOICES,
        verbose_name=_('Type'))
    jurisdictionFipsCode = models.CharField(max_length=5,
        verbose_name=_('FIPS Code'), db_index=True)
    generalRateIntrastate = models.DecimalField(max_digits=8, decimal_places=7,
        verbose_name=_('General Tax Rate - Intrastate'))
    generalRateInterstate = models.DecimalField(max_digits=8, decimal_places=7,
        verbose_name=_('General Tax Rate - Interstate'))
    foodRateIntrastate = models.DecimalField(max_digits=8, decimal_places=7,
        verbose_name=_('Food/Drug Tax Rate - Intrastate'))
    foodRateInterstate = models.DecimalField(max_digits=8, decimal_places=7,
        verbose_name=_('Food/Drug Tax Rate - Interstate'))
    startDate = models.DateField(verbose_name=_('Effective Start Date'))
    endDate = models.DateField(verbose_name=_('Effective End Date'))

    class Meta:
        verbose_name = _("Tax Rate")
        verbose_name_plural = _("Tax Rates")

    def __unicode__(self):
        return u'State %d: Jurisdiction: %s(%s)' % (
            self.state,
            self.jurisdictionFipsCode,
            self.get_jurisdictionType_display(),
        )

    def rate(self, intrastate=False, food=False):
        if intrastate:
            if food:
                return self.foodRateIntrastate
            else:
                return self.generalRateIntrastate
        else:
            if food:
                return self.foodRateInterstate
            else:
                return self.generalRateInterstate

TAX_BOUNDRY_CHOICES = (
    ('Z', 'Zip-5 Record'),
    ('4', 'Zip+4 Record'),
    ('A', 'Address Record'),
)
ODD_EVEN_CHOICES = (
    ('O', 'Odd'),
    ('E', 'Even'),
    ('B', 'Both'),

)
class TaxBoundry(models.Model):
    """
    Records for tax boundries in the default SST format as defined at:
    http://www.streamlinedsalestax.org/Technology/RatesandBoundariesClean082605.pdf
    """
    recordType = models.CharField(max_length=1, choices=TAX_BOUNDRY_CHOICES,
        verbose_name=_('Boundry Type'))
    startDate = models.DateField(verbose_name=_('Effective Start Date'))
    endDate = models.DateField(verbose_name=_('Effective End Date'))
    lowAddress = models.IntegerField(blank=True, null=True,
        verbose_name=_('Low Address Range'))
    highAddress = models.IntegerField(blank=True, null=True,
        verbose_name=_('High Address Range'))
    oddEven = models.CharField(max_length=1, blank=True, null=True, choices=ODD_EVEN_CHOICES,
        verbose_name=_('Odd / Even Range Indicator'))
    streetPreDirection = models.CharField(max_length=2, blank=True, null=True,
        verbose_name=_('State Pre-Directional Abbr.'))
    streetName = models.CharField(max_length=20, blank=True, null=True,
        verbose_name=_('Street Name'))
    streetSuffix = models.CharField(max_length=4, blank=True, null=True,
        verbose_name=_('Street Suffix Abbr.'))
    streetPostDirection = models.CharField(max_length=2, blank=True, null=True,
        verbose_name=_('Street Post Directional'))
    addressSecondaryAbbr = models.CharField(max_length=4, blank=True, null=True,
        verbose_name=_('Address Secondary - Abbr.'))
    addressSecondaryLow = models.IntegerField(blank=True, null=True,
        verbose_name=_('Address Secondary - Low'))
    addressSecondaryHigh = models.IntegerField(max_length=8, blank=True, null=True,
        verbose_name=_('Address Secondary - High'))
    addressSecondaryOddEven = models.CharField(max_length=1, blank=True, null=True,
        choices=ODD_EVEN_CHOICES, verbose_name=_('Address Secondary - Odd/Even'))
    cityName = models.CharField(max_length=28, blank=True, null=True,
        verbose_name=_('City Name'))
    zipCode = models.IntegerField(blank=True, null=True,
        verbose_name=_('Zip Code'))
    plus4 = models.IntegerField(blank=True, null=True,
        verbose_name=_('Zip Code - Plus 4'))
    zipCodeLow = models.IntegerField(blank=True, null=True,
        verbose_name=_('Zip Code - Low'), db_index=True)
    zipExtensionLow = models.IntegerField(blank=True, null=True,
        verbose_name=_('Zip Code Extension - Low'), db_index=True)
    zipCodeHigh = models.IntegerField(blank=True, null=True,
        verbose_name=_('Zip Code - High'), db_index=True)
    zipExtensionHigh = models.IntegerField(blank=True, null=True,
        verbose_name=_('Zip Code Extension - High'), db_index=True)
    serCode = models.CharField(max_length=5, verbose_name=_('Composite SER Code'), blank=True, null=True)
    fipsStateCode = models.CharField(max_length=2, blank=True, null=True,
        verbose_name=_('FIPS State Code'))
    fipsStateIndicator = models.CharField(max_length=2, blank=True, null=True,
        verbose_name=_('FIPS State Indicator'))
    fipsCountyCode = models.CharField(max_length=3, blank=True, null=True,
        verbose_name=_('FIPS County Code'))
    fipsPlaceCode = models.CharField(max_length=5, blank=True, null=True,
        verbose_name=_('FIPS Place Code'))
    fipsPlaceType = models.CharField(max_length=2, blank=True, null=True,
        verbose_name=_('FIPS Place Type'), choices=JURISDICTION_CHOICES)
    special_1_code = models.CharField(max_length=5, verbose_name=_('FIPS Special 1 code'), blank=True, null=True)
    special_1_type = models.CharField(max_length=2, verbose_name=_('FIPS Special 1 type'), blank=True, null=True, choices=JURISDICTION_CHOICES)
    special_2_code = models.CharField(max_length=5, verbose_name=_('FIPS Special 2 code'), blank=True, null=True)
    special_2_type = models.CharField(max_length=2, verbose_name=_('FIPS Special 2 type'), blank=True, null=True, choices=JURISDICTION_CHOICES)
    special_3_code = models.CharField(max_length=5, verbose_name=_('FIPS Special 3 code'), blank=True, null=True)
    special_3_type = models.CharField(max_length=2, verbose_name=_('FIPS Special 3 type'), blank=True, null=True, choices=JURISDICTION_CHOICES)
    special_4_code = models.CharField(max_length=5, verbose_name=_('FIPS Special 4 code'), blank=True, null=True)
    special_4_type = models.CharField(max_length=2, verbose_name=_('FIPS Special 4 type'), blank=True, null=True, choices=JURISDICTION_CHOICES)
    special_5_code = models.CharField(max_length=5, verbose_name=_('FIPS Special 5 code'), blank=True, null=True)
    special_5_type = models.CharField(max_length=2, verbose_name=_('FIPS Special 5 type'), blank=True, null=True, choices=JURISDICTION_CHOICES)
    special_6_code = models.CharField(max_length=5, verbose_name=_('FIPS Special 6 code'), blank=True, null=True)
    special_6_type = models.CharField(max_length=2, verbose_name=_('FIPS Special 6 type'), blank=True, null=True, choices=JURISDICTION_CHOICES)
    special_7_code = models.CharField(max_length=5, verbose_name=_('FIPS Special 7 code'), blank=True, null=True)
    special_7_type = models.CharField(max_length=2, verbose_name=_('FIPS Special 7 type'), blank=True, null=True, choices=JURISDICTION_CHOICES)
    special_8_code = models.CharField(max_length=5, verbose_name=_('FIPS Special 8 code'), blank=True, null=True)
    special_8_type = models.CharField(max_length=2, verbose_name=_('FIPS Special 8 type'), blank=True, null=True, choices=JURISDICTION_CHOICES)
    special_9_code = models.CharField(max_length=5, verbose_name=_('FIPS Special 9 code'), blank=True, null=True)
    special_9_type = models.CharField(max_length=2, verbose_name=_('FIPS Special 9 type'), blank=True, null=True, choices=JURISDICTION_CHOICES)
    special_10_code = models.CharField(max_length=5, verbose_name=_('FIPS Special 10 code'), blank=True, null=True)
    special_10_type = models.CharField(max_length=2, verbose_name=_('FIPS Special 10 type'), blank=True, null=True, choices=JURISDICTION_CHOICES)
    special_11_code = models.CharField(max_length=5, verbose_name=_('FIPS Special 11 code'), blank=True, null=True)
    special_11_type = models.CharField(max_length=2, verbose_name=_('FIPS Special 11 type'), blank=True, null=True, choices=JURISDICTION_CHOICES)
    special_12_code = models.CharField(max_length=5, verbose_name=_('FIPS Special 12 code'), blank=True, null=True)
    special_12_type = models.CharField(max_length=2, verbose_name=_('FIPS Special 12 type'), blank=True, null=True, choices=JURISDICTION_CHOICES)
    special_13_code = models.CharField(max_length=5, verbose_name=_('FIPS Special 13 code'), blank=True, null=True)
    special_13_type = models.CharField(max_length=2, verbose_name=_('FIPS Special 13 type'), blank=True, null=True, choices=JURISDICTION_CHOICES)
    special_14_code = models.CharField(max_length=5, verbose_name=_('FIPS Special 14 code'), blank=True, null=True)
    special_14_type = models.CharField(max_length=2, verbose_name=_('FIPS Special 14 type'), blank=True, null=True, choices=JURISDICTION_CHOICES)
    special_15_code = models.CharField(max_length=5, verbose_name=_('FIPS Special 15 code'), blank=True, null=True)
    special_15_type = models.CharField(max_length=2, verbose_name=_('FIPS Special 15 type'), blank=True, null=True, choices=JURISDICTION_CHOICES)
    special_16_code = models.CharField(max_length=5, verbose_name=_('FIPS Special 16 code'), blank=True, null=True)
    special_16_type = models.CharField(max_length=2, verbose_name=_('FIPS Special 16 type'), blank=True, null=True, choices=JURISDICTION_CHOICES)
    special_17_code = models.CharField(max_length=5, verbose_name=_('FIPS Special 17 code'), blank=True, null=True)
    special_17_type = models.CharField(max_length=2, verbose_name=_('FIPS Special 17 type'), blank=True, null=True, choices=JURISDICTION_CHOICES)
    special_18_code = models.CharField(max_length=5, verbose_name=_('FIPS Special 18 code'), blank=True, null=True)
    special_18_type = models.CharField(max_length=2, verbose_name=_('FIPS Special 18 type'), blank=True, null=True, choices=JURISDICTION_CHOICES)
    special_19_code = models.CharField(max_length=5, verbose_name=_('FIPS Special 19 code'), blank=True, null=True)
    special_19_type = models.CharField(max_length=2, verbose_name=_('FIPS Special 19 type'), blank=True, null=True, choices=JURISDICTION_CHOICES)
    special_20_code = models.CharField(max_length=5, verbose_name=_('FIPS Special 20 code'), blank=True, null=True)
    special_20_type = models.CharField(max_length=2, verbose_name=_('FIPS Special 20 type'), blank=True, null=True, choices=JURISDICTION_CHOICES)

    # Fill in this property to use not-today for looking up the tax rates.
    date = None
    # Set these and we'll use non-default rate.
    useIntrastate = None
    useFood = None

    def get_zip_range(self):
        if self.zipExtensionLow:
            return u'%05d-%04d -> %05d-%04d' % (
                self.zipCodeLow, self.zipExtensionLow, self.zipCodeHigh, self.zipExtensionHigh
            )
        else:
            return u'%05d -> %05d' % (self.zipCodeLow, self.zipCodeHigh)
    zip_range = property(get_zip_range)

    def rates(self, date=None):
        l = list()
        state = self.fipsStateCode
        
        if not date:
            date = _date.today()
            
        # Lookup all the applicable codes.
        for fips in (
            self.fipsStateIndicator, self.fipsCountyCode, self.fipsPlaceCode,
            self.special_1_code, self.special_2_code, self.special_3_code,
            self.special_4_code, self.special_5_code, self.special_6_code,
            self.special_7_code, self.special_8_code, self.special_9_code,
            self.special_10_code, self.special_11_code, self.special_12_code,
            self.special_13_code, self.special_14_code, self.special_15_code,
            self.special_16_code, self.special_17_code, self.special_18_code,
            self.special_19_code, self.special_20_code
        ):
            if not fips:
                continue
            
            rate = TaxRate.objects.get(
                state=state,
                jurisdictionFipsCode=fips,
                startDate__lte=date,
                endDate__gte=date,
            )
            l.append( rate  )
        
        return l

    def get_percentage(self, date=None):
        """
        Emulate being a tax rate by returning a total percentage to tax the customer.
        """
        pct = Decimal('0.00')
        for x in self.rates(date):
            pct += x.rate(intrastate=self.useIntrastate, food=self.useFood)
        return pct
    percentage=property(get_percentage)
    
    def __unicode__(self):
        if self.recordType == 'Z':
            return u'TaxBoundry(Z): %i -- %i' % (
                self.zipCodeLow, self.zipCodeHigh
            )
        elif self.recordType == '4':
            return u'TaxBoundry(4): %i-%i -- %i-%i' % (
                self.zipCodeLow, self.zipExtensionLow,
                self.zipCodeHigh, self.zipExtensionHigh,
            )
        else:
            return u'TaxBoundry(A)'

    @classmethod
    def lookup(cls, zip, ext=None, date=None):
        """Handy function to take a zip code and return the appropriate rates
        for it."""
        if not date:
            date = _date.today()
        
        # Try for a ZIP+4 lookup first if we can.
        if ext:
            try:
                return cls.objects.get(
                    recordType='4',
                    zipCodeLow__lte=zip,
                    zipCodeHigh__gte=zip,
                    zipExtensionLow__lte=ext,
                    zipExtensionHigh__gte=ext,
                    startDate__lte=date,
                    endDate__gte=date,
                )
            except cls.DoesNotExist:
                # Not all zip+4 have entires. That's OK.
                pass
        
        # Try for just the ZIP then.
        try:
            return cls.objects.get(
                recordType='Z',
                zipCodeLow__lte=zip,
                zipCodeHigh__gte=zip,
                startDate__lte=date,
                endDate__gte=date,
            )
        except cls.DoesNotExist:
            return None

    class Meta:
        verbose_name = _("Tax Boundry")
        verbose_name_plural = _("Tax Boundries")


#class TaxCollected(models.Model):
#    order = models.ForeignKey(Order, verbose_name=_("Order"))
#    taxRate = models.ForeignKey(TaxRate, verbose_name=_('Tax Rate'))
#    useIntrastate = models.BooleanField(verbose_name=_('Use Intrastate rate instead of Interstate?'),
#                                        default=True)
#    useFood = models.BooleanField(verbose_name=_('Use food/drug rate instead of general?'),
#                                  default=False)
#
#def save_taxes_collected(order, **kwargs):
#    processor = Processor(order=order)
#    tb = processor.get_boundry()
#

#order_success.connect(save_taxes_colletecd)

import config



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2007-2019 Etienne Robillard <tkadm30@yandex.com>
# All rights reserved.
# See NOTICE file for details.

from functools import wraps

from .django_compat import reverse

__all__ = ('permalink', )

def permalink(view_func):
    """
    Decorator that calls urlresolvers.reverse() to return a URL using
    parameters returned by the decorated function "func".

    "view_func" should be a function that returns a tuple in one of the
    following formats:
        (viewname, viewargs)
        (viewname, viewargs, viewkwargs)
    """
    @wraps(view_func)
    def inner(*args, **kwargs):
        bits = view_func(*args, **kwargs)
        return reverse(bits[0], None, *bits[1:3])
    return inner




"""
ProductShipping  models
""" 
from decimal import Decimal
from django.conf import settings
from django.db import models
from django.utils.translation import get_language, ugettext_lazy as _
from shipping.modules.base import BaseShipper
from product.models import Product
import datetime
import logging
import operator

log = logging.getLogger('shipping.productshipping')

class ProductShippingPriceException(Exception):
    def __init__(self, reason):
        self.reason = reason

class Shipper(BaseShipper):

    def __init__(self, carrier):
        self.id = carrier.key
        self.carrier = carrier
        super(BaseShipper, self).__init__()

    def __str__(self):
        """
        This is mainly helpful for debugging purposes
        """
        return "Shipper: %s" % self.id

    def description(self):
        """
        A basic description that will be displayed to the user when selecting their shipping options
        """
        return self.carrier.description

    def cost(self):
        """
        Complex calculations can be done here as long as the return value is a dollar figure
        """
        assert(self._calculated)
        total = Decimal("0.00")
        for cartitem in self.cart.cartitem_set.all():
            p = cartitem.product
            if p.is_shippable:
                total += self.carrier.price(p) * cartitem.quantity
        return total

    def method(self):
        """
        Describes the actual delivery service (Mail, FedEx, DHL, UPS, etc)
        """
        return self.carrier.method

    def expectedDelivery(self):
        """
        Can be a plain string or complex calcuation returning an actual date
        """
        return self.carrier.delivery

    def valid(self, order=None):
        """
        Can do complex validation about whether or not this option is valid.
        For example, may check to see if the recipient is in an allowed country
        or location.
        """
        if order:
            try:
                for item in order.orderitem_set.all():
                    p = item.product
                    price = self.carrier.price(p)
                            
            except ProductShippingPriceException:
                return False
                
        elif self.cart:
            try:
                price = self.cost()
            
            except ProductShippingPriceException:
                return False
                        
        return True


class Carrier(models.Model):
    key = models.SlugField(_('Key'))
    ordering = models.IntegerField(_('Ordering'), default=0)
    active = models.BooleanField(_('Active'), default=True)

    def _find_translation(self, language_code=None):
        if not language_code:
            language_code = get_language()
            
        c = self.translations.filter(languagecode__exact = language_code)
        ct = c.count()

        if not c or ct == 0:
            pos = language_code.find('-')
            if pos>-1:
                short_code = language_code[:pos]
                log.debug("Carrier: Trying to find root language content for: [%s]", short_code)
                c = self.translations.filter(languagecode__exact = short_code)
                ct = c.count()
                if ct>0:
                    log.debug("Carrier: Found root language content for: [%s]", short_code)

        if not c or ct == 0:
            #log.debug("Trying to find default language content for: %s", self)
            c = self.translations.filter(languagecode__istartswith = settings.LANGUAGE_CODE)
            ct = c.count()

        if not c or ct == 0:
            #log.debug("Trying to find *any* language content for: %s", self)
            c = self.translations.all()
            ct = c.count()

        if ct > 0:
            trans = c[0]
        else:
            trans = None

        return trans

    def delivery(self):
        """Get the delivery, looking up by language code, falling back intelligently.
        """
        trans = self._find_translation()

        if trans:
            return trans.delivery
        else:
            return ""

    delivery = property(delivery)
 
    def description(self):
        """Get the description, looking up by language code, falling back intelligently.
        """
        trans = self._find_translation()

        if trans:
            return trans.description
        else:
            return ""

    description = property(description)
    
    def method(self):
        """Get the description, looking up by language code, falling back intelligently.
        """
        trans = self._find_translation()

        if trans:
            return trans.method
        else:
            return ""

    method = property(method)    
 
    def name(self):
        """Get the name, looking up by language code, falling back intelligently.
        """
        trans = self._find_translation()

        if trans:
            return trans.name
        else:
            return ""

    name = property(name)
    
    def price(self, product):
        """Get a price for this product."""
        prices = product.smart_relation('shipping_price')
        q = prices.filter(carrier=self)
        if q.count() > 0:
            return q[0].price

        else:
            log.debug("No %s shipping price found for %s", self, product.slug)
            raise ProductShippingPriceException('No price available')
            
    def __unicode__(self):
        return u"Carrier: %s [%s]" % (self.name, self.method)
        
    class Meta:
        ordering = ('ordering',)
        
class CarrierTranslation(models.Model):
    carrier = models.ForeignKey('Carrier', related_name='translations')
    languagecode = models.CharField(_('language'), max_length=10, choices=settings.LANGUAGES, )
    name = models.CharField(_('Carrier'), max_length=50, )
    description = models.CharField(_('Description'), max_length=200)
    method = models.CharField(_('Method'), help_text=_("i.e. US Mail"), max_length=200)
    delivery = models.CharField(_('Delivery Days'), max_length=200)
    
    class Meta:
        ordering=('languagecode','name')

class ProductShippingPrice(models.Model):
    product = models.ForeignKey(Product, related_name="shipping_price")
    carrier = models.ForeignKey('Carrier', related_name='tiers')
    price = models.DecimalField(_("Shipping Price"), max_digits=10, decimal_places=2, )
    
    def __unicode__(self):
        return u"ProductShippingPrice: %s" % (self.price)
    
    class Meta:
        ordering = ('carrier','price')

import config



import pygtk
pygtk.require('2.0')
import gtk

from sys import argv

from gazpacho.loader.loader import ObjectBuilder

def on_button1__clicked(button):
    print 'you clicked the button!'

def on_window1__destroy(window):
    gtk.main_quit()

if __name__ == '__main__':
    wt = ObjectBuilder('helloworld.glade')
    wt.signal_autoconnect(globals())
    if '--benchmark' in argv:
        while gtk.events_pending():
            gtk.main_iteration_do()
    else:
        gtk.main()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

from notmm.dbapi.backends.sql import with_session
from notmm.http.decorators import post_required
from notmm.utils.template import direct_to_template

from sandbox_sqlalchemy.config import database as db

#@post_required
@with_session(engine=db.engine)
def test_with_session(request, **kwds):
    session = request.environ['_scoped_session']
    data = {
      'user' : session.query('User').first()
      }
    print data  
    #print session.info
    return direct_to_template(request, template_name='sandbox/test.mako', 
        status=200)
    



from decimal import Decimal, ROUND_HALF_UP, ROUND_DOWN, InvalidOperation, getcontext
from django import forms
from django.utils.translation import ugettext as _
from livesettings import config_value
import logging

log = logging.getLogger('satchmo_utils.numbers')

class RoundedDecimalField(forms.Field):
    def clean(self, value):
        """
        Normalize the field according to cart normalizing rules.
        """
        cartplaces = config_value('SHOP', 'CART_PRECISION')
        roundfactor = config_value('SHOP', 'CART_ROUNDING')    
        
        if not value or value == '':
            value = Decimal(0)

        try:
            value = round_decimal(val=value, places=cartplaces, roundfactor=roundfactor, normalize=True)
        except RoundedDecimalError:
            raise forms.ValidationError(_('%(value)s is not a valid number') % {'value' : value})

        return value

class PositiveRoundedDecimalField(RoundedDecimalField):
    """
    Normalize the field according to cart normalizing rules and force it to be positive.
    """
    def clean(self, value):
        value = super(PositiveRoundedDecimalField, self).clean(value)
        if value<0:
            log.debug('bad val: %s', value)
            raise forms.ValidationError(_('Please enter a positive number'))

        return value

class RoundedDecimalError:
    """
     General Purpose Error Handling used to handle error exceptions
     created in caller.
     Caller name and module taken from
     Activestate Recipe 66062: Determining Current Function Name
        # sys._getframe().f_code.co_name
        # sys._getframe().f_lineno
        # sys._getframe().f_code.co_filename
    """

    def __repr__(self):
        return "RoundedDecimalError - Partial Unit Rounding Error Exception Occured"

    def __init__(self, val='', id='', msg=''):
        import sys
        self.val = val
        self.id = id
        self.msg = msg
        self.caller_name = sys._getframe(1).f_code.co_name		#callers name
        self.caller_module = sys._getframe(1).f_code.co_filename	#module name of caller
        self.caller_lineno = sys._getframe(1).f_lineno		#line number of caller

def round_decimal(val='0', places=None, roundfactor='0', normalize=True):
    """
    PARTIAL UNIT ROUNDING DECIMAL
    Converts a valid float, integer, or string to a decimal number with a specified
    number of decimal places, performs "partial unit rounding", and decimal normalization.

    METHOD ARGUEMENTS:
    `val` The value to be converted and optionally formated to decimal.
    `places` The decimal precision is defined by integer "places" and
        must be <= the precision defined in the decimal.Decimal context.
    `roundfactor` (partial unit rounding factor) If purf is between -1 and 1, purf rounds up
        (positive purf value) or down (negative purf value) in factional "purf" increments.
    `normalize` If normalize is True (any value other than False), then rightmost zeros are truncated.
    """

    #-- Edit function arguments and set necessary defaults

    if str(normalize) == 'False': normalize = False     #Allow templates to submit False from filter

    try:
        roundfactor = Decimal(str(roundfactor))
    except (InvalidOperation, ValueError):
        raise RoundedDecimalError(val=roundfactor, id =1, msg="roundfactor - InvalidOperation or ValueError")  #reraise exception and return to caller
    if not (abs(roundfactor) >= 0 and abs(roundfactor) <= 1):
        raise RoundedDecimalError(val=roundfactor, id=2, msg="roundfactor - Out of Range - must be between -1 and +1")
    try:
        if places != None: places = int(places)
    except ValueError:
        raise RoundedDecimalError(val=places, id=3, msg='ValueError, Invalid Integer ')
    if places > getcontext().prec:
        raise RoundedDecimalError(val=places, id=4, msg='places Exceeds Decimal Context Precision')
    try:
        decval =Decimal(str(val))
    except (InvalidOperation, UnicodeEncodeError):
        raise RoundedDecimalError(val=val, id=5, msg='InvalidOperation - val cannot be converted to Decimal')
    
    #-- Round decimal number by the Partial Unit Rounding Factor
    if roundfactor and decval%roundfactor:
        if roundfactor < 0: roundby = 0
        else: roundby = (decval/abs(decval))*roundfactor	#change sign of roudby to decval
        decval=(decval//roundfactor*roundfactor)+roundby #round up or down by next roundfactor increment
    
    #-- Adjust number of decimal places if caller provided decimal places
    if places != None:
        decmask = '0.'.ljust(places+2,'0') #i.e. => '.00' if places eq 2
        decval=decval.quantize(Decimal(decmask), rounding=ROUND_DOWN)  #convert to Decimal and truncate to two decimals
    
    #-- normalize - strips the rightmost zeros... i.e. 2.0 => returns as 2
    if normalize:
        # if the number has no decimal portion return just the number with no decimal places
        # if the number has decimal places (i.e. 3.20), normalize the number (to 3.2)
        # This check is necesary because normalizing a number which trails in zeros (i.e. 200 or 200.00) normalizes to
        # scientific notation (2E+2)
        if decval==decval.to_integral():
            decval = decval.quantize(Decimal('1'))
        else:
            decval.normalize()
    
    return decval
    
def trunc_decimal(val, places):
    """Legacy compatibility, rounds the way the old satchmo 0.8.1 used to round."""
    if val is None or val == '':
       return val
    if val < 0:
        roundfactor = "-0.01"
    else:
        roundfactor = "0.01"
    return round_decimal(val=val, places=places, roundfactor=roundfactor, normalize=True)



from satchmo_ext.wishlist.models import ProductWish
from django.contrib import admin
from django.utils.translation import get_language, ugettext_lazy as _


class ProductWishOptions(admin.ModelAdmin):
    list_display = ('contact', 'product', 'create_date')
    ordering = ('contact', '-create_date', 'product')

admin.site.register(ProductWish, ProductWishOptions)





from django import forms
from django.utils.translation import ugettext, ugettext_lazy as _

from app_plugins.models import PluginPoint, Plugin, LABEL_RE

def validate_label(value):
    if not LABEL_RE.search(value):
        raise forms.ValidationError(ugettext(
            u"This value must contain only letters, "
            u"numbers, underscores, and '.' dots."
        ))
    else:
        return value

class AdminPluginPointForm(forms.ModelForm):
    class Meta:
        model = PluginPoint
    
    def clean_label(self):
        value = self.cleaned_data["label"]
        return validate_label(value)

class AdminPluginForm(forms.ModelForm):
    class Meta:
        model = Plugin
    
    def clean_label(self):
        value = self.cleaned_data["label"]
        return validate_label(value)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2007-2019 Etienne Robillard <tkadm30@yandex.com>
# All rights reserved.
"""Decorator functions to interact with Schevo databases"""

from functools import wraps
from notmm.controllers.schevo import SchevoController
from notmm.controllers.zodb import ZODBController


__all__ = ('with_schevo_database', 'with_durus_database')

def with_schevo_database(dbname, controller_class=ZODBController):
    """
    Decorator that adds a Schevo database object reference
    in the ``request.environ`` dictionary.

    """
    def decorator(view_func, **kwargs):
        @wraps(view_func, **kwargs)
        def _wrapper(req, *args, **kwargs):
            wsgi_app = controller_class(req, dbname)
            req.environ[wsgi_app.environ_key] = wsgi_app.db
            response = view_func(req, **kwargs)
            return response
        return _wrapper
    return decorator

def with_durus_database(dbname, controller_class=SchevoController):
    return with_schevo_database(dbname, controller_class=controller_class)



"""Taken and modified from the dbsettings project.

http://code.google.com/p/django-values/
"""
from decimal import Decimal
from django import forms
from django.contrib.sites.models import Site
from django.core.exceptions import ImproperlyConfigured
try:
    from django.utils import simplejson
except ImportError:
    import simplejson
from collections import OrderedDict
from django.utils.encoding import smart_str
from django.utils.translation import gettext, ugettext_lazy as _
from keyedcache import cache_set
from livesettings.models import find_setting, LongSetting, Setting, SettingNotSet
from livesettings.overrides import get_overrides
from satchmo_utils import load_module, is_string_like, is_list_or_tuple
import datetime
import logging
import signals

__all__ = ['BASE_GROUP', 'ConfigurationGroup', 'Value', 'BooleanValue', 'DecimalValue', 'DurationValue',
      'FloatValue', 'IntegerValue', 'ModuleValue', 'PercentValue', 'PositiveIntegerValue', 'SortedDotDict',
      'StringValue', 'LongStringValue', 'MultipleStringValue']

_WARN = {}

log = logging.getLogger('configuration')

NOTSET = object()

class SortedDotDict(OrderedDict):

    def __getattr__(self, key):
        try:
            return self[key]
        except:
            raise AttributeError, key

    def __iter__(self):
        vals = self.values()
        for k in vals:
            yield k
            
    def values(self):
        vals = super(SortedDotDict, self).values()
        vals = [v for v in vals if isinstance(v, (ConfigurationGroup, Value))]
        vals.sort()
        return vals

class ConfigurationGroup(SortedDotDict):
    """A simple wrapper for a group of configuration values"""
    def __init__(self, key, name, *args, **kwargs):
        """Create a new ConfigurationGroup.
        
        Arguments:
        - key
        - group name - for display to user
        
        Named Arguments:
        - ordering: integer, optional, defaults to 1.
        - requires: See `Value` requires.  The default `requires` all member values will have if not overridden.
        - requiresvalue: See `Values` requires_value.  The default `requires_value` if not overridden on the `Value` objects.
        """
        self.key = key
        self.name = name
        self.ordering = kwargs.pop('ordering', 1)
        self.requires = kwargs.pop('requires', None)
        if self.requires:
            reqval = kwargs.pop('requiresvalue', key)
            if not is_list_or_tuple(reqval):
                reqval = (reqval, reqval)
                
            self.requires_value = reqval[0]
            self.requires.add_choice(reqval)
            
        super(ConfigurationGroup, self).__init__(*args, **kwargs)
                
    def __cmp__(self, other):
        return cmp((self.ordering, self.name), (other.ordering, other.name))
        
    def __eq__(self, other):
        return (type(self) == type(other) 
                and self.ordering == other.ordering 
                and self.name == other.name)

    def __ne__(self, other):
        return not self == other
        
    def values(self):
        vals = super(ConfigurationGroup, self).values()
        return [v for v in vals if v.enabled()]

BASE_GROUP = ConfigurationGroup('BASE', _('Base Settings'), ordering=0)

class Value(object):

    creation_counter = 0

    def __init__(self, group, key, **kwargs):
        """
        Create a new Value object for configuration.
        
        Args:
            - `ConfigurationGroup`
            - key - a string key
            
        Named arguments:
            - `description` - Will be passed to the field for form usage.  Should be a translation proxy.  Ex: _('example')
            - `help_text` - Will be passed to the field for form usage.
            - `choices` - If given, then the form field will use a select box
            - `ordering` - Defaults to alphabetical by key if not given.
            - `requires` - If given as a `Value`, then this field will only be rendered if that Value evaluates true (for Boolean requires) or the proper key is in the associated value.
            - `requiresvalue` - If set, then this field will only be rendered if that value is in the list returned by self.value. Defaults to self.key.
            - `hidden` - If true, then render a hidden field.
            - `default` - If given, then this Value will return that default whenever it has no assocated `Setting`.
        """
        self.group = group
        self.key = key
        self.description = kwargs.get('description', None)
        self.help_text = kwargs.get('help_text')
        self.choices = kwargs.get('choices',[])
        self.ordering = kwargs.pop('ordering', 0)
        self.hidden = kwargs.pop('hidden', False)
        
        self.requires = kwargs.pop('requires', None)
        if self.requires:
            reqval = kwargs.pop('requiresvalue', key)
            if not is_list_or_tuple(reqval):
                reqval = (reqval, reqval)
                
            self.requires_value = reqval[0]
            self.requires.add_choice(reqval)
            
        elif group.requires:
            self.requires = group.requires
            self.requires_value = group.requires_value
        
        if kwargs.has_key('default'):
            self.default = kwargs.pop('default')
            self.use_default = True                
        else:
            self.use_default = False
        
        self.creation_counter = Value.creation_counter
        Value.creation_counter += 1

    def __cmp__(self, other):
        return cmp((self.ordering, self.description, self.creation_counter), (other.ordering, other.description, other.creation_counter))
        
    def __eq__(self, other):
        if type(self) == type(other):
            return self.value == other.value
        else:
            return self.value == other
            
    def __iter__(self):
        return iter(self.value)
        
    def __unicode__(self):
        return unicode(self.value)
        
    def __str__(self):
        return str(self.value)
        
    def add_choice(self, choice):
        """Add a choice if it doesn't already exist."""
        if not is_list_or_tuple(choice):
            choice = (choice, choice)
        skip = False
        for k, v in self.choices:
            if k == choice[0]:
                skip = True
                break
        if not skip:
            self.choices += (choice, )
        
    def choice_field(self, **kwargs):
        if self.hidden:
            kwargs['widget'] = forms.MultipleHiddenInput()
        return forms.ChoiceField(choices=self.choices, **kwargs)

    def _choice_values(self):
        choices = self.choices
        vals = self.value
        return [x for x in choices if x[0] in vals]

    choice_values = property(fget=_choice_values)

    def copy(self):
        new_value = self.__class__(self.key)
        new_value.__dict__ = self.__dict__.copy()
        return new_value
        
    def _default_text(self):
        if not self.use_default:
            note = ""
        else:
            if self.default == "":
                note = _('Default value: ""')
            
            elif self.choices:
                work = []
                for x in self.choices:
                    if x[0] in self.default:
                        work.append(smart_str(x[1]))
                note = gettext('Default value: ') + ", ".join(work)
        
            else:
                note = _("Default value: %s") % unicode(self.default)
        
        return note
        
    default_text = property(fget=_default_text)
        
    def enabled(self):
        enabled = False
        try:
            if not self.requires:
                enabled = True
            else:
                v = self.requires.value
                if self.requires.choices:
                    enabled = self.requires_value == v or self.requires_value in v
                elif v:
                    enabled = True
        except SettingNotSet:
            pass
        return enabled
        
    def make_field(self, **kwargs):
        if self.choices:
            if self.hidden:
                kwargs['widget'] = forms.MultipleHiddenInput()
            field = self.choice_field(**kwargs)
        else:
            if self.hidden:
                kwargs['widget'] = forms.HiddenInput()
            field = self.field(**kwargs)
            
        field.group = self.group
        field.default_text = self.default_text
        return field

    def make_setting(self, db_value):
        log.debug('new setting %s.%s', self.group.key, self.key)
        return Setting(group=self.group.key, key=self.key, value=db_value)
        
    def _setting(self):
        return find_setting(self.group.key, self.key)
        
    setting = property(fget = _setting)

    def _value(self):            
        try:
            val = self.setting.value
            
        except SettingNotSet, sns:
            if self.use_default:
                val = self.default
            else:
                val = NOTSET
                                
        except AttributeError, ae:
            log.error("Attribute error: %s", ae)
            log.error("%s: Could not get _value of %s", self.key, self.setting)
            raise(ae)
            
        except Exception, e:
            global _WARN
            log.error(e)
            if str(e).find("configuration_setting") > -1:
                if not _WARN.has_key('configuration_setting'):
                    log.warn('Error loading setting %s.%s from table, OK if you are in syncdb', self.group.key, self.key)
                    _WARN['configuration_setting'] = True
                    
                if self.use_default:
                    val = self.default
                else:
                    raise ImproperlyConfigured("All settings used in startup must have defaults, %s.%s does not", self.group.key, self.key)
            else:
                import traceback
                traceback.print_exc()
                log.warn("Problem finding settings %s.%s, %s", self.group.key, self.key, e)
                raise SettingNotSet("Startup error, couldn't load %s.%s" %(self.group.key, self.key))
        return val

    def update(self, value):    
        use_db, overrides = get_overrides()
        
        if use_db:
            current_value = self.value

            new_value = self.to_python(value)
            if current_value != new_value:        
                db_value = self.get_db_prep_save(value)
                try:
                    s = self.setting
                    s.value = db_value
                
                except SettingNotSet:
                    s = self.make_setting(db_value)

                if self.use_default and self.default == new_value:
                    if s.id:
                        log.info("Deleted setting %s.%s", self.group.key, self.key)
                        s.delete()
                else:
                    log.info("Updated setting %s.%s = %s", self.group.key, self.key, value)
                    s.save()
                
                signals.configuration_value_changed.send(self, old_value=current_value, new_value=new_value, setting=self)
            
                return True
        else:
            log.debug('not updating setting %s.%s - livesettings db is disabled',self.group.key, self.key)
            
        return False

    def value(self):
        val = self._value()
        return self.to_python(val)

    value = property(fget = value)
        
    def editor_value(self):
        val = self._value()
        return self.to_editor(val)

    editor_value = property(fget = editor_value)

    # Subclasses should override the following methods where applicable

    def to_python(self, value):
        "Returns a native Python object suitable for immediate use"
        if value == NOTSET:
            value = None
        return value

    def get_db_prep_save(self, value):
        "Returns a value suitable for storage into a CharField"
        if value == NOTSET:
            value = ""
        return unicode(value)

    def to_editor(self, value):
        "Returns a value suitable for display in a form widget"
        if value == NOTSET:
            return NOTSET
        return unicode(value)

###############
# VALUE TYPES #
###############

class BooleanValue(Value):

    class field(forms.BooleanField):

        def __init__(self, *args, **kwargs):
            kwargs['required'] = False
            forms.BooleanField.__init__(self, *args, **kwargs)

    def add_choice(self, choice):
        # ignore choice adding for boolean types
        pass

    def to_python(self, value):
        if value in (True, 't', 'True', 1, '1'):
            return True
        return False
        
    to_editor = to_python

class DecimalValue(Value):
    class field(forms.DecimalField):

           def __init__(self, *args, **kwargs):
               kwargs['required'] = False
               forms.DecimalField.__init__(self, *args, **kwargs)

    def to_python(self, value):
        if value==NOTSET:
            return Decimal("0")
            
        try:
            return Decimal(value)
        except TypeError, te:
            log.warning("Can't convert %s to Decimal for settings %s.%s", value, self.group.key, self.key)
            raise TypeError(te)
        
    def to_editor(self, value):
        if value == NOTSET:
            return "0"
        else:
            return unicode(value)

# DurationValue has a lot of duplication and ugliness because of issue #2443
# Until DurationField is sorted out, this has to do some extra work
class DurationValue(Value):

    class field(forms.CharField):
        def clean(self, value):
            try:
                return datetime.timedelta(seconds=float(value))
            except (ValueError, TypeError):
                raise forms.ValidationError('This value must be a real number.')
            except OverflowError:
                raise forms.ValidationError('The maximum allowed value is %s' % datetime.timedelta.max)

    def to_python(self, value):
        if value == NOTSET:
            value = 0
        if isinstance(value, datetime.timedelta):
            return value
        try:
            return datetime.timedelta(seconds=float(value))
        except (ValueError, TypeError):
            raise forms.ValidationError('This value must be a real number.')
        except OverflowError:
            raise forms.ValidationError('The maximum allowed value is %s' % datetime.timedelta.max)

    def get_db_prep_save(self, value):
        if value == NOTSET:
            return NOTSET
        else:
            return unicode(value.days * 24 * 3600 + value.seconds + float(value.microseconds) / 1000000)

class FloatValue(Value):
    
    class field(forms.FloatField):

        def __init__(self, *args, **kwargs):
            kwargs['required'] = False
            forms.FloatField.__init__(self, *args, **kwargs)

    def to_python(self, value):
        if value == NOTSET:
            value = 0
        return float(value)
        
    def to_editor(self, value):
        if value == NOTSET:
            return "0"
        else:
            return unicode(value)
            
class IntegerValue(Value):
    class field(forms.IntegerField):

        def __init__(self, *args, **kwargs):
            kwargs['required'] = False
            forms.IntegerField.__init__(self, *args, **kwargs)

    def to_python(self, value):
        if value == NOTSET:
            value = 0
        return int(value)

    def to_editor(self, value):
        if value == NOTSET:
            return "0"
        else:
            return unicode(value)


class PercentValue(Value):

    class field(forms.DecimalField):

        def __init__(self, *args, **kwargs):
            kwargs['required'] = False
            forms.DecimalField.__init__(self, 100, 0, 5, 2, *args, **kwargs)

        class widget(forms.TextInput):
            def render(self, *args, **kwargs):
                # Place a percent sign after a smaller text field
                attrs = kwargs.pop('attrs', {})
                attrs['size'] = attrs['max_length'] = 6
                return forms.TextInput.render(self, attrs=attrs, *args, **kwargs) + '%'

    def to_python(self, value):
        if value == NOTSET:
            value = 0
        return Decimal(value) / 100

    def to_editor(self, value):
        if value == NOTSET:
            return "0"
        else:
            return unicode(value)

class PositiveIntegerValue(IntegerValue):

    class field(forms.IntegerField):

        def __init__(self, *args, **kwargs):
            kwargs['min_value'] = 0
            forms.IntegerField.__init__(self, *args, **kwargs)


class StringValue(Value):
    
    class field(forms.CharField):
        def __init__(self, *args, **kwargs):
            kwargs['required'] = False
            forms.CharField.__init__(self, *args, **kwargs)
            
    def to_python(self, value):
        if value == NOTSET:
            value = ""
        return unicode(value)
            
    to_editor = to_python

class LongStringValue(Value):

    class field(forms.CharField):
        def __init__(self, *args, **kwargs):
            kwargs['required'] = False
            kwargs['widget'] = forms.Textarea()
            forms.CharField.__init__(self, *args, **kwargs)

    def make_setting(self, db_value):
        log.debug('new long setting %s.%s', self.group.key, self.key)
        return LongSetting(group=self.group.key, key=self.key, value=db_value)        

    def to_python(self, value):
        if value == NOTSET:
            value = ""
        return unicode(value)

    to_editor = to_python


class MultipleStringValue(Value):

    class field(forms.CharField):

        def __init__(self, *args, **kwargs):
            kwargs['required'] = False
            forms.CharField.__init__(self, *args, **kwargs)
    
    def choice_field(self, **kwargs):
        kwargs['required'] = False
        return forms.MultipleChoiceField(choices=self.choices, **kwargs)
                
    def get_db_prep_save(self, value):
        if is_string_like(value):
            value = [value]
        return simplejson.dumps(value)
            
    def to_python(self, value):
        if not value or value == NOTSET:
            return []
        if is_list_or_tuple(value):
            return value
        else:
            try:
                return simplejson.loads(value)
            except:
                if is_string_like(value):
                    return [value]
                else:
                    log.warning('Could not decode returning empty list: %s', value)
                    return []
               
             
    to_editor = to_python
            
class ModuleValue(Value):
    """Handles setting modules, storing them as strings in the db."""
    
    class field(forms.CharField):

        def __init__(self, *args, **kwargs):
            kwargs['required'] = False
            forms.CharField.__init__(self, *args, **kwargs)

    def load_module(self, module):
        """Load a child module"""
        value = self._value()
        if value == NOTSET:
            raise SettingNotSet("%s.%s", self.group.key, self.key)
        else:
            return load_module("%s.%s" % (value, module))
            
    def to_python(self, value):
        if value == NOTSET:
            v = {}            
        else:
            v = load_module(value)
        return v
            
    def to_editor(self, value):
        if value == NOTSET:
            value = ""
        return value    




#!/usr/bin/env python
from notmm.utils.configparse import setup_all
#import logging_conf

global_conf = {}

setup_all('sandbox_sqlalchemy', global_conf)
setup_all('template', global_conf)
app_conf = global_conf['sandbox_sqlalchemy']




from django.utils.translation import ugettext_lazy as _
from livesettings import *
from satchmo_store.shop.config import SHOP_GROUP

config_register(
    StringValue(SHOP_GROUP,
        'WISHLIST_SLUG',
        description=_("Wishlist slug"),
        help_text=_("The url slug for wishlists.  Requires server restart if changed."),
        default="wishlist"))



"""
Handle a Purchase Order payments.
"""
from django.utils.translation import ugettext as _
from payment.modules.base import BasePaymentProcessor, ProcessorResult, NOTSET

class PaymentProcessor(BasePaymentProcessor):

    def __init__(self, settings):
        super(PaymentProcessor, self).__init__('purchaseorder', settings)

    def can_refund(self):
        return True

    def prepare_data(self, order):
        super(PaymentProcessor, self).prepare_data(order)

    def capture_payment(self, testing=False, order=None, amount=NOTSET):
        """
        Purchase Orders are always successful.
        """
        if not order:
            order = self.order

        if amount == NOTSET:
            amount = order.balance

        payment = self.record_payment(order=order, amount=amount, 
            transaction_id="PO", reason_code='0')

        return ProcessorResult(self.key, True, _('Success'), payment)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
import sys
PY3K = sys.version_info[0] == 3
import os
import posixpath
import glob

# see bootstrap for installing thoses requirements
# using easy_install (err Distribute)
try:
    import yaml 
except ImportError as e:
    raise ImportError("Please run the ./bootstrap script first.")

workdir = os.getcwd()
admindir = posixpath.join(workdir, 'admin')

# adds 'lib' to sys.path 
#sys.path.insert(0, posixpath.join(workdir, 'lib'))
#import notmm
#DEVELOPMENT_RELEASE = False
#buildName = notmm.getBuildName(release=isRelease)

from setuptools import setup, find_packages, Extension
#from pkg_resources import resource_stream

# Do import buildutils commands if available!
try:
    import buildutils
except ImportError:
    print('Consider installing the buildutils module!')

libs = find_packages(where='lib')

# Meta info in YAML data format for improved readability
meta = yaml.load(open(posixpath.join(admindir, 'PKG-INFO.in')))

scripts_data = glob.glob('tools/schevo-*') + ['tools/httpserver.py']

staticdir = posixpath.abspath(os.path.join(admindir, 'static'))

classifiers = [
    (str(item)) for item in open(posixpath.abspath(os.path.join(staticdir, \
    'classifiers.txt'))).read().split('\n') if item is not '']

setup(
    name=meta['Name'],
    version=meta['Version'],
    description=meta['Summary'], 
    long_description=meta['Description'],
    author=meta['Author'],
    author_email=meta['Author-email'],
    license=meta['License'],
    keywords=meta['Keywords'],
    url=meta['Homepage'],
    #maintainer=meta['Maintainer'],
    #maintainer_email=meta['Maintainer-email'],
    scripts=scripts_data,

    # Include stuff which belong in SVN or mentioned in MANIFEST.in
    include_package_data=True,

    # Location where packages lives
    package_dir={'': 'lib'},
    packages=libs,

    # Package classifiers are read from static/classifiers.txt
    classifiers=classifiers,

    # Add Cython compiled extensions
    #cmdclass={'build_ext': build_ext},
    #ext_modules=ext_modules,
    
    # Minimal packages required when doing `python setup.py install`.
    install_requires=[
        #'Django>=1.11.17',  # Maintainers should not need this :-)
        'pytz>=2017.2',     # django dependency (optional)
        'Beaker>=1.9.0',    # memcached support (optional)
        'configobj>=4.7.2', # in notmm.utils.configparse (required)
        'argparse>=1.1',    # Used by tools/httpserver.py (required)
        'demjson>=2.2.4',   # JSON support (optional)
        'Mako>=1.0.7',      # Mako template backend (optional)
        #'feedparser>=5.1.2'# RSS 2.0 parsing (optional)
        #'docutils>=0.8.1',         # Docutils support (optional)
        #'python-epoll>=1.0',       # For epoll support Linux only (optional)
        #'pytidylib>=0.2.1',        # PyTidyLib support      (optional) 
        #'python-memcached>=1.58',  # memcached support      (optional)
        'Werkzeug>=0.14.1',         # Werkzeug  support      (optional) 
        #'gevent==1.4.0',            # Gevent    support      (optional)
        'ZODB>=5.3.0',              # ZODB backend support   (optional)
        'ZEO>=5.1.0',               # ZEO backend support    (optional)
        #'Elixir>=0.7.1',            # Elixir support         (optional)
        'libschevo>=4.1',         # Schevo backend support (required)
        'blogengine2>=0.9.6',       # BlogEngine 2.0 support (optional)
        'uWSGI>=2.0.18',          # uWSGI support          (required)
        'django-hotsauce-oauthclient>=0.3.1', # Experimental OAuth 2.0 support
        'transaction>=2.4',         # transaction support    (required)
        'persistent>=4.4.3'         # persistence support    (required)
    ],
    zip_safe=False
)




from satchmo_ext.newsletter import update_subscription
from satchmo_store.contact.forms import ContactInfoForm
from signals_ahoy.signals import collect_urls, form_postsave
import logging

log = logging.getLogger('newsletter.listeners')

def contact_form_listener(sender, object=None, formdata=None, form=None, **kwargs):
    if 'newsletter' not in formdata:
        subscribed = False
    else:
        subscribed = formdata['newsletter']
    
    log.debug('Updating newletter subscription for %s to %s', object, subscribed)
    update_subscription(object, subscribed)
    
def start_listening():
    from urls import add_newsletter_urls
    from satchmo_store import shop
    
    form_postsave.connect(contact_form_listener, sender=ContactInfoForm)
    collect_urls.connect(add_newsletter_urls, sender=shop)




"""
This code is heavily based on samples found here -
http://www.b-list.org/weblog/2006/06/14/django-tips-template-context-processors

It is used to add some common variables to all the templates
"""
from django.conf import settings as site_settings
from product.models import Category, Discount
from satchmo_store.shop import get_satchmo_setting
from satchmo_store.shop.models import Config, Cart
from satchmo_store.shop.signals import satchmo_context
from satchmo_utils import current_media_url, request_is_secure
import logging

log = logging.getLogger('shop_context')

def settings(request):
    shop_config = Config.objects.get_current()
    cart = Cart.objects.from_request(request)

    all_categories = Category.objects.by_site()

    try:
        sale = Discount.objects.get_sale()
    except Discount.DoesNotExist:
        sale = None

    ctx = {
        'shop_base': get_satchmo_setting('SHOP_BASE'),
        'shop' : shop_config,
        'shop_name': shop_config.store_name,
        'media_url': current_media_url(request),
        'cart_count': cart.numItems,
        'cart': cart,
        'categories': all_categories,
        'is_secure' : request_is_secure(request),
        'request' : request,
        'login_url': site_settings.LOGIN_URL,
        'logout_url': site_settings.LOGOUT_URL,
        'sale': sale
    }

    satchmo_context.send(shop_config, context=ctx)

    return ctx



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from notmm.utils.urlmap import RegexURLMap, url
from notmm.utils.template import direct_to_template

# a custom/generic login view
handleLogin = 'blogengine.contrib.session.login'
handleLogout = 'blogengine.contrib.session.logout'

urlpatterns = RegexURLMap()
urlpatterns.add_routes('sandbox.views',
        url(r'^$|index.html$', 'index', {
            'template_name': 'sandbox/test.mako',
            'charset': 'utf-8' # must specify mimetype or charset
        }),
)
# Add the authentication views
urlpatterns.add_routes('',
        # Custom authentication backend wrapper (authkit based)
        # see nomainapp..controllers.auth.LoginController
        url(r'^session_login/$', handleLogin, \
            dict(template_name='auth/login.mako')),
        url(r'^session_logout/$', handleLogout, \
            dict(template_name='auth/logout.mako')),
)

urlpatterns.include('sandbox.config.extras')

#blogengine api v1
#urlpatterns.include('blogengine.contrib.api_v1.urls', prefix="^blog/")





from django.core.management.base import NoArgsCommand
from django.core.management import call_command

class Command(NoArgsCommand):
    help = "Load Satchmo country (l10n) data."
    
    def handle_noargs(self, **options):
        """Load l10n fixtures"""
        call_command('loaddata', 'us_tax.yaml', interactive=True)



# Copyright (C) 2005 by SICEm S.L. and Async Open Source
# Copyright (C) 2005 Red Hat, Inc.
# Copyright (C) 2005,2006 Async Open Source
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import gtk

from gazpacho.properties import prop_registry, TransparentProperty, EnumType, \
     CustomProperty
from gazpacho.widgets.base.base import ContainerAdaptor, ObjectDataProp

class WindowAdaptor(ContainerAdaptor):
    def create(self, context, interactive=True):
        window = super(WindowAdaptor, self).create(context, interactive)
        window.set_default_size(440, 250)
        return window

class TypeHintProperty(CustomProperty, EnumType):
    def get(self):
        value = self.object.get_data('gazpacho::type-hint')
        return value or gtk.gdk.WINDOW_TYPE_HINT_NORMAL

    def set(self, value):
        self.object.set_data('gazpacho::type-hint', value)

prop_registry.override_simple('GtkWindow::modal', ObjectDataProp,
                              qdata='gazpacho::modal')

prop_registry.override_simple('GtkWindow::type-hint', TypeHintProperty)

prop_registry.override_simple('GtkWindow::default-width',
                              minimum=-1, maximum=10000)
prop_registry.override_simple('GtkWindow::default-height',
                              minimum=-1, maximum=10000)
prop_registry.override_simple('GtkWindow::type', TransparentProperty,
                              editable=False)
prop_registry.override_simple('GtkWindow::role', translatable=False)



from django import template
import decimal

register = template.Library()

@register.filter
def flatten_discounts(items):
    flat_items = []
    for item in items:
        if item.unit_price != 0:
            f = item.discount / item.unit_price
        else:
            f = decimal.Decimal('0')
        free_qty = f.quantize(decimal.Decimal('1'), rounding=decimal.ROUND_FLOOR)
        partial_qty = int(not (f % decimal.Decimal('1')) == decimal.Decimal('0')) # bool to int
        full_qty = item.quantity - (free_qty + partial_qty)
        if free_qty:
            flat_items.append({'orderitem': item, 'price': 0, 'quantity': free_qty})
        if partial_qty:
            partial_price = item.unit_price - item.discount + item.unit_price * free_qty
            flat_items.append({'orderitem': item, 'price': partial_price, 'quantity': partial_qty})
        if full_qty:
            flat_items.append({'orderitem': item, 'price': item.unit_price, 'quantity': full_qty})
    return flat_items



# Copyright (C) 2004,2005 by SICEm S.L. and Imendio AB
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

class CommandManager(object):
    """
    This class is the entry point accesing the commands.
    Every undoable action in Gazpacho is wrapped into a command.
    The stack of un/redoable actions is stored in the Project class
    so each project is independent on terms of undo/redo.
    """

    def __init__(self):
        self._in_progress = False

    def undo(self, project):
        """Undo the last command if there is such a command"""
        if project.undo_stack.has_undo():
            cmd = project.undo_stack.pop_undo()
            cmd.undo()

    def redo(self, project):
        """Redo the last undo command if there is such a command"""
        if project.undo_stack.has_redo():
            cmd = project.undo_stack.pop_redo()
            cmd.redo()

    def execute(self, cmd, project, nested=True):
        self._in_progress = True
        ret = cmd.execute()

        # If nested is set to true and we're already executing another
        # command, don't add it to the undo/redo list
        # This is a big hack, but we need a transactional like API to
        # solve it properly
        if nested or not self._in_progress:
            project.undo_stack.push_undo(cmd)
        self._in_progress = True
        return ret

command_manager = CommandManager()
